#!/usr/bin/python3
# Fit and draw BNMR data
# Derek Fujimoto
# November 2017

from tkinter import *
from tkinter import ttk, filedialog, messagebox
from bdata import bdata, bmerged
from scipy.optimize import curve_fit

try:
    import matplotlib as mpl
except ImportError:
    pass

try:
    from mpl_toolkits.mplot3d import Axes3D
except ImportError as errmsg:
    print('No 3D axes drawing available')
    print(errmsg)

import sys, os, datetime, textwrap, yaml
import webbrowser, subprocess, importlib, logging, warnings
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import bdata as bd
import weakref as wref

from multiprocessing import Queue
import queue

from bfit import __version__, logger_name, icon_path
from bfit.gui.tab_fileviewer import fileviewer
from bfit.gui.tab_fetch_files import fetch_files
from bfit.gui.tab_fit_files import fit_files
from bfit.gui.calculator_nqr_B0 import calculator_nqr_B0
from bfit.gui.calculator_nqr_B0_hh6 import calculator_nqr_B0_hh6
from bfit.gui.calculator_nmr_B1 import calculator_nmr_B1
from bfit.gui.calculator_nmr_atten import calculator_nmr_atten
from bfit.gui.popup_drawstyle import popup_drawstyle
from bfit.gui.popup_deadtime import popup_deadtime
from bfit.gui.popup_redraw_period import popup_redraw_period
from bfit.gui.popup_terminal import popup_terminal
from bfit.gui.popup_units import popup_units
from bfit.gui.popup_set_ppm_reference import popup_set_ppm_reference
from bfit.gui.popup_ongoing_process import popup_ongoing_process
from bfit.gui.popup_set_histograms import popup_set_histograms
from bfit.backend.PltTracker import PltTracker
from bfit.backend.fitdata import fitdata
import bfit.backend.colors as colors

# interactive plotting
plt.ion()

# filter warnings related to new dkeys on read
warnings.simplefilter('ignore', bd.exceptions.DkeyWarning)


__doc__="""
    BNMR/BNQR data visualization and curve fitting.

    Hotkeys:

        General
            Command-------------Effect
            ctrl+n:             set draw mode "new"
            ctrl+s:             set draw mode "stack"
            ctrl+r:             set draw mode "redraw"

        File Details
            Command-------------Effect
            return:             fetch file
            ctrl+return:        draw file
            shift+return:       draw file
            ctrl+a:             toggle check all
        Fit
            Fetch Data
                Command---------Focus-----------------------Effect
                return:         run/year entry:             fetch
                                SLR rebin:                  set checked
                                bin omit entry (checked):   set checked
                ctrl+return:                                draw file
                shift+return:                               draw file

            Fit Data

        View Fit Results

    Derek Fujimoto
    November 2017
    """

# =========================================================================== #
class bfit(object):
    """
        Build the mainframe and set up the runloop for the tkinter GUI.

        Data Fields:
            asym_dict_keys: asym calc and draw types
            bnmr_data_dir:  string, directory for bnmr data
            bnqr_data_dir:  string, directory for bnqr data
            correct_bkgd:   BooleanVar, if true apply slr background correction
            data:           dict of fitdata objects for drawing/fitting, keyed by run #
            deadtime:       float, value of deadtime in s or scaling for local calcs
            deadtime_switch:BooleanVar, if true, use deadtime correction
            deadtime_global:BooleanVar, if true, deadtime value is dt, else is scaling
            draw_style:     StringVar, draw window types # stack, redraw, new
            draw_components:list of titles for labels, options to export, draw.
            draw_fit:       BooleanVar, if true draw fits after fitting
            draw_ppm:       BoolVar for drawing as ppm shift
            draw_prebin:    BoolVar, if true draw prebeam bins
            draw_rel_peak0: BoolVar for drawing frequencies relative to peak0
            draw_standardized_res: BoolVar for drawing residuals as standardized
            norm_with_param:BoolVar, if true estimate normalization from data only
            hist_select:    histogram selection for asym calcs (blank for defaults)
            label_default:  StringVar() name of label defaults for fetch
            logger:         logging object
            logger_name:    string of unique logger name
            mainframe:      main frame for the object
            menus:          dict {title: Menu} of menubar options
            minimizer:      StringVar: path to python module with fitter object
            notebook:       contains all tabs for operations:
                fileviewer
                fetch_files
                fit_files
            plt:            PltTracker for tracking figures
            ppm_reference:  reference freq in Hz for ppm calulations
            probe_species:  StringVar() name of probe species, bdata.life key.
            root:           tkinter root instance
            rounding:       number of decimal places to round results to in display
            routine_mod:    module with fitting routines
            style:          dict, drawing styles
            thermo_channel: StringVar for tracking how temperature is calculated
            units:          dict:(float, str). conversion rate from original to display units
            update_period:  int, update spacing in s.
            use_nbm:        BooleanVar, use NBM in asym calculations

    """
    bnmr_archive_label = "BNMR_ARCHIVE"
    bnqr_archive_label = "BNQR_ARCHIVE"
    rounding = 4       # number of significant figures to round results to in display
    norm_alph_diff_time = 0.1   # number of seconds to take average over when
                                # normalizing alpha diffusion runs
    legend_max_draw = 8 # max number of items to draw before removing the legend

    # csymmetry calculation options
    asym_dict_keys = {'20':["Combined Helicity",
                            "Split Helicity",
                            "Combined Normalized",
                            "Matched Helicity",
                            "Histograms",
                            "Positive Helicity",
                            "Negative Helicity",
                            ],
                      '1f':["Combined Helicity",
                            "Split Helicity",
                            "Raw Scans",
                            "Shifted Split",
                            "Shifted Combined",
                            "Combined Normalized",
                            "Histograms",
                            "Positive Helicity",
                            "Negative Helicity",
                            ],
                      '1x':["Combined Helicity",
                            "Split Helicity",
                            "Raw Scans",
                            "Shifted Split",
                            "Shifted Combined",
                            "Combined Normalized",
                            "Histograms",
                            "Positive Helicity",
                            "Negative Helicity",
                            ],
                      '1n':["Combined Helicity",
                            "Split Helicity",
                            "Raw Scans",
                            "Matched Peak Finding",
                            "Histograms",
                            "Positive Helicity",
                            "Negative Helicity",
                            ],
                      '1e':["Combined Helicity",
                            "Split Helicity",
                            "Raw Scans",
                            "Histograms",
                            "Positive Helicity",
                            "Negative Helicity",
                            ],
                      '1d':["Combined Helicity",
                            "Split Helicity",
                            "Raw Scans",
                            "Histograms",
                            "Positive Helicity",
                            "Negative Helicity",
                            ],
                      '1c':["Combined Helicity",
                            "Split Helicity",
                            "Raw Scans",
                            "Histograms",
                            "Positive Helicity",
                            "Negative Helicity",
                            ],
                      '1w':["Combined Helicity",
                            "Split Helicity",
                            "Raw Scans",
                            "Shifted Split",
                            "Shifted Combined",
                            "Combined Normalized",
                            "Histograms",
                            "Positive Helicity",
                            "Negative Helicity",
                            ],
                      '2e':["Combined Hel Slopes",
                            "Combined Hel Diff",
                            "Combined Hel Raw",
                            "Split Hel Slopes",
                            "Split Hel Diff",
                            "Split Hel Raw",
                            "Split Slopes Shifted",
                            "Split Diff Shifted",
                            "Split Raw Shifted",
                            ],
                      '2h':["Combined Helicity",
                            "Split Helicity",
                            "Combined Normalized",
                            "Positive Helicity",
                            "Alpha Diffusion",
                            "Alpha Diff Normalized",
                            "Combined Hel (Alpha Tag)",
                            "Split Hel (Alpha Tag)",
                            "Combined Hel (!Alpha Tag)",
                            "Split Hel (!Alpha Tag)",
                            "Histograms",
                            "Negative Helicity",
                            "Matched Helicity",
                            ]}

    # asymmetry calculation codes
    asym_dict = {"Combined Helicity"        :'c',
                 "Split Helicity"           :'h',
                 "Positive Helicity"        :'p',
                 "Negative Helicity"        :'n',
                 "Forward Counter"          :'fc',
                 "Backward Counter"         :'bc',
                 "Right Counter"            :'rc',
                 "Left Counter"             :'lc',
                 "Matched Helicity"         :'hm',
                 "Shifted Split"            :'hs',
                 "Shifted Combined"         :'cs',
                 "Combined Normalized"      :'cn',
                 "Matched Peak Finding"     :'hp',
                 "Raw Scans"                :'r',
                 "Histograms"               :'rhist',
                 "Combined Hel Raw"         :'raw_c',
                 "Combined Hel Slopes"      :'sl_c',
                 "Combined Hel Diff"        :'dif_c',
                 "Split Hel Raw"            :'raw_h',
                 "Split Hel Slopes"         :'sl_h',
                 "Split Hel Diff"           :'dif_h',
                 "Split Raw Shifted"        :'raw_hs',
                 "Split Slopes Shifted"     :'sl_hs',
                 "Split Diff Shifted"       :'dif_hs',
                 "Alpha Diffusion"          :'ad',
                 "Alpha Diff Normalized"    :'adn',
                 "Combined Hel (Alpha Tag)" :"at_c",
                 "Split Hel (Alpha Tag)"    :"at_h",
                 "Combined Hel (!Alpha Tag)":"nat_c",
                 "Split Hel (!Alpha Tag)"   :"nat_h",
                 }

    # valid thermometer channels to read from
    thermo_keys = ('A', 'B', '(A+B)/2')

    # draw axis labels
    xlabel_dict={'20':"Time (%s)",
                 '2h':"Time (%s)",
                 '2e':'Frequency (%s)',
                 '1f':'Frequency (%s)',
                 '1x':'Frequency (%s)',
                 '1w':'x Parameter (%s)',
                 '1e':'Field (G)',
                 '1d':'Laser Power',
                 '1c':'Camp Variable',
                 '1n':'Voltage (%s)'}

    ylabel_dict={'ad':r'$N_\alpha~/~N_\beta$', # otherwise, label as Asymmetry
                 'adn':r'$N_\alpha~~N_\beta$',
                 'hs':r'$\mathcal{A}~-~\mathcal{A}(\nu_\mathrm{max}$)',
                 'cs':r'$\mathcal{A}~-~\mathcal{A}(\nu_\mathrm{max}$)',
                 'csf':r'$\mathcal{A}~-$ Baseline',
                 'cn1':r'$\mathcal{A}~/~\mathcal{A}(\nu_\mathrm{max}$)',
                 'cn1f':r'$\mathcal{A}$ / Baseline',
                 'cn2':r'$\mathcal{A}~/~\mathcal{A}(t_\mathrm{min}$)',
                 'cn2f':r'$\mathcal{A}$ / Amplitude',
                 'rhist':'Counts'}

    # histogram names for x axis
    x_tag={'20':"time_s",
           '2h':"time_s",
           '2e':"time",
           '1f':'freq',
           '1x':'freq',
           '1w':'xpar',
           '1d':'las',
           '1e':'mA',
           '1c':'camp',
           '1n':'mV'}

    # minimizers
    minimizers = {'curve_fit (trf)':'bfit.fitting.fitter_curve_fit',
                  'migrad (hesse)':'bfit.fitting.fitter_migrad_hesse',
                  'migrad (minos)':'bfit.fitting.fitter_migrad_minos',
                  }

    # define draw componeents in draw_param and labels
    draw_components = ('Temperature (K)', '1000/T (1/K)', 'Impl. Energy (keV)',
                       'Platform Bias (kV)', 'Run Number', 'B0 Field (T)',
                       'Unique Id', 'Sample', 'RF Level DAC', 'Chi-Squared',
                       'Run Duration (s)', 'Start Time', 'Title', 'Year',
                       'Cryo Lift Set (mm)', 'Cryo Lift Read (mm)',
                       'He Mass Flow', 'CryoEx Mass Flow', 'Needle Set (turns)',
                       'Needle Read (turns)', 'Laser Power (V)', 'Laser Wavelength (nm)',
                       'Laser Wavenumber (1/cm)', 'Target Bias (kV)', 'NBM Rate (count/s)',
                       'Sample Rate (count/s)', 'NBM Counts', 'Sample Counts')

    # ======================================================================= #
    def __init__(self, testfn=None, commandline=False):
        """
            testfn: if not none, expect a function handle with input self to run
                    automate setting parameters, button pushes, etc for rapid
                    testing
            commandline:    if True leave user in interactive mode with no mainloop running
        """
        # logging
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing v%s' % __version__ + '-'*50)
        self.logger.info('bdata: v%s' % bd.__version__)

        # default settings
        self.update_period = 10  # s
        self.ppm_reference = 41270000 # Hz
        self.hist_select = ''    # histogram selection for asym calculations
        self.use_nbm_settings = {'default':False,
                                 '1n':True}
        self.style = {'linestyle':'None',
                      'linewidth':mpl.rcParams['lines.linewidth'],
                      'marker':'.',
                      'markersize':mpl.rcParams['lines.markersize'],
                      'capsize':0.,
                      'elinewidth':mpl.rcParams['lines.linewidth'],
                      'alpha':1.,
                      'fillstyle':'full'}

        # units: mode:[conversion rate from original to display units, unit]
        self.units = {  '1f':[1e-6, 'MHz'],
                        '1x':[1e-6, 'MHz'],
                        '2e':[1e-6, 'MHz'],
                        '1w':[1, 'Hz'],
                        '1n':[0.001, 'V'],
                        '1e':[0.001, 'A'],
                        '20':[1, 's'],
                        '2h':[1, 's'],
                        }

        # for fitdata objects
        self.data = {}

        # set data directories
        try:
            self.bnmr_data_dir = os.environ[self.bnmr_archive_label]
            self.bnqr_data_dir = os.environ[self.bnqr_archive_label]
        except(AttributeError, KeyError):
            self.bnmr_data_dir = os.getcwd()
            self.bnqr_data_dir = os.getcwd()

        # plot tracker
        self.plt = PltTracker()

        # root
        root = Tk()
        self.root = root
        root.title("bfit: β-NMR and β-NQR Data Analysis "+\
                   "(version %s)" % __version__)
        root.columnconfigure(0, weight=1)
        root.rowconfigure(0, weight=1)

        # styling
        root.option_add('*tearOff', FALSE)
        root.option_add("*Font", colors.font)
        root.option_add("*Background",          colors.background)
        root.option_add("*DisabledBackground",  colors.background)
        root.option_add("*ReadonlyBackground",  colors.readonly)
        root.option_add("*Borderwidth", 2)

        # don't change all foregrounds or you will break the filedialog windows
        root.option_add("*Menu*Foreground",     colors.foreground)
        root.option_add("*Spinbox*Foreground",  colors.foreground)
        root.option_add("*Listbox*Foreground",  colors.foreground)
        root.option_add("*Text*Foreground",     colors.foreground)
        root.option_add("*Scrollbar.Background",colors.foreground)

        root.option_add("*Entry.Foreground",    colors.insertbackground)
        root.option_add("*Entry.Background",    colors.fieldbackground)
        root.option_add("*Entry.HighlightBackground",colors.background)
        root.option_add("*Entry.DisabledBackground",colors.entry_disabled)

        ttk_style = ttk.Style()
        ttk_style.configure('.',   font=colors.font,
                                   background=colors.background,
                                   foreground=colors.foreground,
                                   arrowcolor=colors.foreground,
                                   borderwidth=2)

        ttk_style.configure("TEntry", foreground=colors.foreground,
                                      fieldbackground=colors.fieldbackground)

        ttk_style.map('.', background=[('disabled', colors.background)],
                           fieldbackground=[('selected', colors.selected)])

        ttk_style.configure('TNotebook.Tab', padding=[50, 2])
        ttk_style.configure("TNotebook.Tab", background=colors.background)
        ttk_style.map("TNotebook.Tab", background=[("selected", colors.tab)])

        ttk_style.map("TCheckbutton", foreground=[('selected', colors.selected),
                                                 ('disabled', colors.disabled)],
                                      indicatorcolor=[('selected', colors.selected)])
        ttk_style.map("TRadiobutton", foreground=[('selected', colors.selected),
                                                  ('disabled', colors.foreground)],
                                      indicatorcolor=[('selected', colors.selected)])
        ttk_style.map('TCombobox', fieldbackground=[('readonly', colors.background)])

        ttk_style.configure('TSpinbox', borderwidth=0, background=colors.background)
        ttk_style.map('TSpinbox', borderwidth=[('selected', 1)])

        ttk_style.configure('TProgressbar',
                            borderwidth=1,
                            background=colors.background)

        # icon
        self.set_icon(root)

        # key bindings
        root.bind('<Return>', self.return_binder)
        root.bind('<KP_Enter>', self.return_binder)
        root.bind('<Control-Key-Return>', self.draw_binder)
        root.bind('<Control-Key-KP_Enter>', self.draw_binder)
        root.bind('<Shift-Key-Return>', self.draw_binder)
        root.bind('<Shift-Key-KP_Enter>', self.draw_binder)

        root.bind('<Control-Key-n>', self.set_style_new)
        root.bind('<Control-Key-s>', self.set_style_stack)
        root.bind('<Control-Key-r>', self.set_style_redraw)
        root.bind('<Control-Key-a>', self.set_check_all)

        root.bind('<Control-Key-1>', lambda x: self.set_focus_tab(idn=0))
        root.bind('<Control-Key-2>', lambda x: self.set_focus_tab(idn=1))
        root.bind('<Control-Key-3>', lambda x: self.set_focus_tab(idn=2))

        root.bind("<Button-4>", self.scroll_binder)
        root.bind("<Button-5>", self.scroll_binder)

        root.bind("<Control-Key-o>", lambda x: self.load_state())

        # event bindings
        root.protocol("WM_DELETE_WINDOW", self.on_closing)

        # main frame
        mainframe = ttk.Frame(root, pad=5)
        mainframe.grid(column=0, row=0, sticky=(N, W, E, S))
        mainframe.columnconfigure(0, weight=1)
        mainframe.rowconfigure(0, weight=1)

        # nbm variables
        self.nbm_dict = {''  :BooleanVar(),
                         '1n':BooleanVar()}

        self.nbm_dict[''].set(False)
        self.nbm_dict['1n'].set(True)

        # deadtime switches
        self.deadtime_switch = BooleanVar()
        self.deadtime_switch.set(False)
        self.deadtime_global = BooleanVar()
        self.deadtime_global.set(True)
        self.deadtime = 0 # deadtime in s, or neg asymmetry scaling for deadtime
                          # calculations, depending on the value of self.global_deadtime


        # Menu bar options ----------------------------------------------------
        root.option_add('*tearOff', FALSE)
        menubar = Menu(root)
        root['menu'] = menubar

        # File
        menu_file = Menu(menubar, title='File')
        menu_file.add_command(label='Search archive', command=self.search_archive)
        menu_file.add_command(label='Run Commands', command=lambda:popup_terminal(wref.proxy(self)))
        menu_file.add_command(label='Export Data', command=self.do_export)
        menu_file.add_command(label='Export Fits', command=self.do_export_fit)
        menu_file.add_command(label='Save State', command=self.save_state)
        menu_file.add_command(label='Load State', command=self.load_state)
        menu_file.add_command(label='Close All Figures', command=self.do_close_all)
        menu_file.add_command(label='Exit', command=sys.exit)
        menubar.add_cascade(menu=menu_file, label='File')

        # Settings
        menu_settings = Menu(menubar, title='Settings')
        menubar.add_cascade(menu=menu_settings, label='Settings')
        menu_settings_dir = Menu(menu_settings)
        menu_settings_lab = Menu(menu_settings)
        menu_settings_probe = Menu(menu_settings, selectcolor=colors.selected)
        menu_settings_thermo = Menu(menu_settings, selectcolor=colors.selected)

        # Settings cascade commands
        menu_settings.add_cascade(menu=menu_settings_dir, label='Data directory')
        menu_settings.add_command(label='Drawing style',
                command=self.set_draw_style)
        menu_settings.add_command(label='Histograms',
                command=self.set_histograms)
        menu_settings.add_cascade(menu=menu_settings_lab, label='Labels default')
        menu_settings.add_command(label='PPM Reference Frequecy',
                command=self.set_ppm_reference)
        menu_settings.add_cascade(menu=menu_settings_probe, label='Probe Species')
        menu_settings.add_command(label='Redraw period',
                command=self.set_redraw_period)
        menu_settings.add_command(label="System matplotlibrc",
                command=self.set_matplotlib)
        menu_settings.add_cascade(menu=menu_settings_thermo, label='Thermometer Channel')
        menu_settings.add_command(label="Units",
                command=self.set_units)

        # Settings: data directory
        menu_settings_dir.add_command(label="β-NMR", command=self.set_bnmr_dir)
        menu_settings_dir.add_command(label="β-NQR", command=self.set_bnqr_dir)

        # Settings: set label default
        self.label_default = StringVar()
        self.label_default.set('Unique Id')
        for dc in sorted(self.draw_components):
            menu_settings_lab.add_radiobutton(label=dc,
                variable=self.label_default, value=dc, command=self.set_all_labels,
                selectcolor=colors.selected)

        # Settings: set probe species
        self.probe_species = StringVar()
        self.probe_species.set('Li8')
        lifekeys = list(bd.life.keys())
        lifekeys.sort()
        for k in lifekeys:
            if 'err' not in k:
                menu_settings_probe.add_radiobutton(label=k,
                        variable=self.probe_species,
                        value=k,
                        command=self.set_probe_species)

        # Settings: set thermometer channel
        self.thermo_channel = StringVar()
        self.thermo_channel.set(self.thermo_keys[0])
        for k in self.thermo_keys:
            menu_settings_thermo.add_radiobutton(label=k,
                    variable=self.thermo_channel,
                    value=k,
                    command=self.set_thermo_channel)

        # calculate
        menu_calculate = Menu(menubar, title='Calculate')
        menubar.add_cascade(menu=menu_calculate, label='Calculate')
        menu_calculate.add_command(label='Deadtime correction',
                command=self.set_deadtime)
        menu_calculate.add_command(label='NQR B0 (HH3)', command=calculator_nqr_B0)
        menu_calculate.add_command(label='NQR B0 (HH6)', command=calculator_nqr_B0_hh6)
        menu_calculate.add_command(label='NMR B1', command=calculator_nmr_B1)
        menu_calculate.add_command(label='NMR B1 Attenuation', command=calculator_nmr_atten)

        # Draw style
        self.draw_style = StringVar()
        self.draw_style.set("stack")
        self.draw_ppm = BooleanVar()
        self.draw_ppm.set(False)
        self.draw_rel_peak0 = BooleanVar()
        self.draw_rel_peak0.set(False)
        self.draw_standardized_res = BooleanVar()
        self.draw_standardized_res.set(True)
        self.use_nbm = self.nbm_dict['']
        self.norm_with_param = BooleanVar()
        self.norm_with_param.set(True)
        self.draw_fit = BooleanVar()
        self.draw_fit.set(True)
        self.draw_prebin = BooleanVar()
        self.draw_prebin.set(False)
        self.correct_bkgd = BooleanVar()
        self.correct_bkgd.set(True)

        menu_draw = Menu(menubar, title='Draw Mode')
        menubar.add_cascade(menu=menu_draw, label='Draw Mode')
        menu_draw.add_radiobutton(label="Draw in new window",
                variable=self.draw_style, value='new', underline=8,
                selectcolor=colors.selected)
        menu_draw.add_radiobutton(label="Stack in existing window",
                variable=self.draw_style, value='stack', underline=0,
                selectcolor=colors.selected)
        menu_draw.add_radiobutton(label="Redraw in existing window",
                variable=self.draw_style, value='redraw', underline=0,
                selectcolor=colors.selected)

        menu_draw.add_separator()
        menu_draw.add_checkbutton(label="Draw after fitting",
                variable=self.draw_fit, selectcolor=colors.selected)
        menu_draw.add_checkbutton(label="Normalize with fit results",
                variable=self.norm_with_param, selectcolor=colors.selected)
        menu_draw.add_checkbutton(label="Draw residuals as standardized",
                variable=self.draw_standardized_res, selectcolor=colors.selected)
        menu_draw.add_checkbutton(label="Draw 1f/1x as PPM shift",
                variable=self.draw_ppm, selectcolor=colors.selected,
                command = lambda : self.set_1f_shift_style('ppm'))
        menu_draw.add_checkbutton(label="Draw 1f/1x relative to peak_0",
                variable = self.draw_rel_peak0, selectcolor = colors.selected,
                command = lambda : self.set_1f_shift_style('peak'))
        menu_draw.add_checkbutton(label="Draw prebeam bins",
                variable=self.draw_prebin, selectcolor=colors.selected)

        menu_draw.add_separator()
        menu_draw.add_checkbutton(label="Use NBM in asymmetry", \
                variable=self.nbm_dict[''], selectcolor=colors.selected)
        menu_draw.add_checkbutton(label="SLR background correction", \
                variable=self.correct_bkgd, selectcolor=colors.selected)

        # Fitting minimizers
        menu_mini = Menu(menubar, title='Minimizer')
        menubar.add_cascade(menu=menu_mini, label='Minimizer')

        self.minimizer = StringVar()
        self.minimizer.set(list(self.minimizers.values())[0])
        for k, m in self.minimizers.items():
            menu_mini.add_radiobutton(label=k, \
                    variable=self.minimizer,
                    value=m,
                    selectcolor=colors.selected,
                    command=self.set_fit_routine)
        menu_mini.add_checkbutton(label='Other', \
                variable=self.minimizer,
                selectcolor=colors.selected,
                command=self.set_fit_routine_with_popup)

        # Help
        menu_help = Menu(menubar, title='Help')
        menubar.add_cascade(menu=menu_help, label='Help')
        menu_help.add_command(label='Show help wiki', command=self.help)
        menu_help.add_command(label='Cite bfit', command=self.citation)
        menu_help.add_command(label='Update bfit', command=self.update_bfit)
        menu_help.add_command(label="What's new?", command=self.whatsnew)
        menu_help.add_command(label='Report an issue', command=self.report_issue)

        # load default fitting routines
        self.routine_mod = importlib.import_module(self.minimizer.get())

        # Top Notebook: File Viewer, Fit, Fit Viewer -------------------------
        noteframe = ttk.Frame(mainframe, relief='sunken', pad=5)
        notebook = ttk.Notebook(noteframe)
        file_viewer_tab = ttk.Frame(notebook)
        fetch_files_tab = ttk.Frame(notebook)
        fit_files_tab = ttk.Frame(notebook)

        notebook.add(file_viewer_tab, text='Inspect')
        notebook.add(fetch_files_tab, text=' Fetch ')
        notebook.add(fit_files_tab,  text='  Fit  ')

        # set drawing styles
        notebook.bind("<<NotebookTabChanged>>", self.set_tab_change)

        # gridding
        notebook.grid(column=0, row=0, sticky=(N, E, W, S))
        noteframe.grid(column=0, row=0, sticky=(N, E, W, S))
        noteframe.columnconfigure(0, weight=1)
        noteframe.rowconfigure(0, weight=1)

        # Notetabs
        self.fileviewer = fileviewer(file_viewer_tab, self)
        self.fetch_files = fetch_files(fetch_files_tab, self)
        self.fit_files = fit_files(fit_files_tab, self)

        # set instance variables ---------------------------------------------
        self.mainframe = mainframe
        self.notebook = notebook

        # save menus
        self.menus = {'menubar': menubar}
        for child in menubar.winfo_children():
            self.menus[child['title']] = child

        # testing
        if testfn is not None:
            testfn(self)

        # logging
        self.logger.debug('Initialization success. Starting mainloop.')

        # runloop
        if commandline:
            root.update_idletasks()
            root.update()
        else:
            self.root.mainloop()

    # ======================================================================= #
    def __del__(self):
        if hasattr(self, 'fileviewer'):  del self.fileviewer
        if hasattr(self, 'fetch_files'): del self.fetch_files
        if hasattr(self, 'fitviewer'):   del self.fitviewer

        try:
            plt.close('all')
        except ImportError:
            pass

    # ======================================================================= #
    def citation(self):
        """Display publication page"""
        self.logger.info('Opening https://joss.theoj.org/papers/10.21105/joss.03598')
        webbrowser.open('https://joss.theoj.org/papers/10.21105/joss.03598')

    # ======================================================================= #
    def do_close_all(self):
        """Close all open figures"""
        plt.close('all')
        for k in self.plt.plots:    self.plt.plots[k] = []
        for k in self.plt.active:   self.plt.active[k] = 0

    # ======================================================================= #
    def do_export(self):
        """Export selected files to csv format. Calls the appropriate function
        depending on what tab is selected. """

        idx = self.notebook.index('current')
        self.logger.debug('Exporting for notebook index %d', idx)
        if idx == 0:        # data viewer
            self.fileviewer.export()
        elif idx == 1:        # data fetch_files
            self.fetch_files.export()
        elif idx == 2:        # fit viewer
            self.fetch_files.export()
        else:
            pass

    # ======================================================================= #
    def do_export_fit(self): self.fit_files.export_fit()

    # ======================================================================= #
    def draw_binder(self, *args):
        """
            Switch between various functions of the shift+enter button.
            Bound to ctrl+enter
        """

        idx = self.notebook.index('current')
        self.logger.debug('Drawing for notebook index %d', idx)
        if idx == 0:        # data viewer
            self.fileviewer.draw(figstyle='inspect')
        elif idx == 1:        # data fetch_files
            self.fetch_files.draw_all(figstyle='data')
        elif idx == 2:        # fit viewer
            self.fit_files.draw_param()
        else:
            pass

    # ======================================================================= #
    def export(self, data, filename, rebin=1, omit=''):
        """Export single data file as csv"""

        self.logger.info('Exporting single run (%d) as "%s"', data.run, filename)

        # settings
        title_dict = {'c':"combined",
                      'p':"positive_helicity",
                      'n':"negative_helicity",
                      'f':'forward_counter',
                      'b':'backward_counter',
                      'r':'right_counter',
                      'l':'left_counter',
                      'time_s':'time_s',
                      'freq':"freq_Hz",
                      'mV':'voltage_mV',
                      'xpar':'x_parameter',
                      }

        index_list = ['time_s', 'freq_Hz', 'voltage_mV', 'x_parameter']

        # get asymmetry
        asym = data.asym(hist_select=self.hist_select, omit=omit,
                         rebin=rebin, nbm=self.use_nbm.get(),
                         slr_bkgd_corr=self.correct_bkgd.get())

        # get new keys
        asym_out = {}
        for k in asym.keys():
            if k == 'custom':
                asym_out[data.ppg.customv_name_read.units] = asym[k]
            elif len(asym[k]) == 2:
                asym_out[title_dict[k]] = asym[k][0]
                asym_out[title_dict[k]+"_err"] = asym[k][1]
            else:
                asym_out[title_dict[k]] = asym[k]

        # make pandas dataframe
        df = pd.DataFrame.from_dict(asym_out)

        # set index
        if 'custom' in asym.keys():
            df.set_index(data.ppg.customv_name_read.units, inplace=True)
        else:
            for i in index_list:
                if i in asym_out.keys():
                    df.set_index(i, inplace=True)
                    break

        # make header
        header = [  '# %s' % data.id,
                    '# %s' % data.title,
                    '# Rebin: %d' % rebin,
                    '# Alternate histogram selection string: "%s"' % self.hist_select,
                    '# Data computed from NBM: %s' % self.use_nbm.get()]

        header_foot=['# ',
                    '# Generated by bfit v%s on %s' % (__version__, datetime.datetime.now()),
                    '#']

        if data.mode in ('20', '00', '2h'):
            header = [  *header,
                        '# Background corrected: %s' % self.correct_bkgd.get(),
                        *header_foot
                     ]
        else:
            header = [  *header,
                        '# Bin omission string: "%s"' % omit,
                        *header_foot
                     ]

        # write to file
        with open(filename, 'w') as fid:
            fid.write('\n'.join(header) + '\n')

        try:
            df.to_csv(filename, mode='a+')
        except AttributeError:
            self.logger.exception('Export file write failed')
            pass

    # ======================================================================= #
    def get_asym_mode(self, obj):
        """ Get asymmetry calculation type"""
        id_string = obj.asym_type.get()
        return self.asym_dict[id_string]

    # ======================================================================= #
    def get_label(self, data):
        """ Get label for plot
            Input: fitdata object.
        """

        # the thing to switch on
        select = self.label_default.get()
        self.logger.debug('Fetching plot label for "%s" (run %d)', select, data.run)

        # Data file options
        try:
            if select == 'Temperature (K)':
                label = "%d K" % int(round(data.temperature.mean))

            elif select == 'B0 Field (T)':
                if data.field > 0.1:
                    label = "%3.2f T"% np.around(data.field, 2)  # field (Tesla)
                else:
                    label = "%3.2f G" % np.round(data.field*1e4)# field (Gauss)

            elif select == 'RF Level DAC':
                label = str(int(data.bd.camp.rf_dac.mean))

            elif select == 'Platform Bias (kV)':
                label = "%.2f kV" % np.around(data.bias, 2)

            elif select == 'Impl. Energy (keV)':
                label = "%.2f keV" % np.around(data.bd.beam_keV, 2)

            elif select == 'Run Duration (s)':
                label = "%d s" % int(data.bd.duration)

            elif select == 'Run Number':
                label = str(data.run)

            elif select == 'Sample':
                label = data.bd.sample

            elif select == 'Start Time':
                label = data.bd.start_date

            elif select == 'Title':
                label = data.bd.title

            elif select == '1000/T (1/K)':
                label = '%3.3f 1/K' % np.around(1000/data.temperature.mean, 2)

            elif select == 'Chi-Squared':
                try:
                    label = "%.2f" % np.around(data.chi)
                except(KeyError, AttributeError):
                    label = ""

            elif select == 'Year':
                label = '%d' % data.year

            elif select == 'Unique Id':
                label = '%s' % data.id

            elif 'Cryo Lift Set (mm)' in select:
                label = '%3.2f mm' % np.around(data.bd.camp.clift_set.mean, 2)

            elif 'Cryo Lift Read (mm)' in select:
                label = '%3.2f mm' % np.around(data.bd.camp.clift_read.mean, 2)

            elif 'He Mass Flow' in select:
                var = 'mass_read' if data.area == 'BNMR' else 'he_read'
                label = '%3.2f' % np.around(data.bd.camp[var].mean, 2)

            elif 'CryoEx Mass Flow' in select:
                label = '%3.2f' % np.around(data.bd.camp.cryo_read.mean, 2)

            elif 'Needle Set (turns)' in select:
                label = '%3.2f turns' % np.around(data.bd.camp.needle_set.mean, 2)

            elif 'Needle Read (turns)' in select:
                label = '%3.2f turns' % np.around(data.bd.camp.needle_pos.mean, 2)

            elif 'Laser Power (V)' in select:
                label = '%3.2f' % np.around(data.bd.epics.las_pwr.mean, 2)

            elif 'Laser Wavelength (nm)' in select:
                label = '%3.2f' % np.around(data.bd.epics.las_lambda.mean, 2)

            elif 'Laser Wavenumber (1/cm)' in select:
                label = '%3.5f' % np.around(data.bd.epics.las_wavenum.mean, 5)

            elif 'Target Bias (kV)' in select:
                label = '%3.2f kV' % np.around(data.bd.epics.target_bias.mean, 2)

            elif 'NBM Rate (count/s)' in select:
                rate = np.sum([data.hist['NBM'+h].data \
                               for h in ('F+', 'F-', 'B-', 'B+')])/data.duration
                label = '%3.2f count/s' % np.around(rate, 2)

            elif 'Sample Rate (count/s)' in select:
                hist = ('F+', 'F-', 'B-', 'B+') if data.area == 'BNMR' \
                                             else ('L+', 'L-', 'R-', 'R+')

                rate = np.sum([data.hist[h].data for h in hist])/data.duration
                label = '%3.2f count/s' % np.around(rate, 2)

            else:
                label = str(data.run)
        except Exception as err:
            print(str(err))
            label = '%d (Error)' % data.run

        return label

    # ======================================================================= #
    def get_latest_year(self):
        """Get the year which has the last data set in it."""

        # get the current year
        year = datetime.datetime.now().year

        # get paths
        try:
            nmr_path = os.environ[self.bnmr_archive_label]
            nqr_path = os.environ[self.bnqr_archive_label]
        except KeyError:
            nmr_path = os.path.join(bd._mud_data, 'bnmr')
            nqr_path = os.path.join(bd._mud_data, 'bnqr')

        # functions to check for data (NMR or NQR)
        no_nmr = lambda y: not os.path.isdir(os.path.join(nmr_path, str(y)))
        no_nqr = lambda y: not os.path.isdir(os.path.join(nqr_path, str(y)))

        # check data
        while (no_nmr(year) and no_nqr(year)) and year > 0:
            year -= 1

        self.logger.debug('Latest year with data: %d (NMR: %s, NQR: %s)',
                          year, not no_nmr(year), not no_nqr(year))
        return year

    # ======================================================================= #
    def get_run_key(self, data=None, r=-1, y=-1):
        """For indexing data dictionary"""

        if type(data) is bdata:
            return '.'.join(map(str, (data.year, data.run)))
        elif type(data) is bmerged:
            runs = textwrap.wrap(str(data.run), 5)
            years = textwrap.wrap(str(data.year), 4)
            return '+'.join(['%s.%s' % (y, r) for y, r in zip(years, runs)])
        elif type(data) is fitdata:
            return data.id
        elif r>0 and y>0:
            return '.'.join(map(str, (y, r)))
        else:
            raise RuntimeError('Bad run key input')

    # ======================================================================= #
    def help(self):
        """Display help wiki"""
        self.logger.info('Opening help wiki')
        webbrowser.open('https://github.com/dfujim/bfit/wiki')

    # ======================================================================= #
    def load_state(self, filename=None):
        """
            Load the state of the gui
        """

        # get the filename
        if filename is None:
            filename = filedialog.askopenfilename(filetypes=[('yaml', '*.yaml'),
                                                             ('allfiles', '*')])
            if not filename:
                return

        self.logger.info('Loading program state from %s', filename)

        # load the object with the data
        with open(filename, 'r') as fid:
            from_file = yaml.safe_load(fid)

        # clear loaded runs
        fetch_tab = self.fetch_files
        fetch_tab.remove_all()

        # bfit parameters
        self.style = from_file['style']
        self.hist_select = from_file['hist_select']
        self.draw_style.set(from_file['draw_style'])
        self.draw_fit.set(from_file['draw_fit'])

        self.probe_species.set(from_file['probe_species'])
        self.set_probe_species()

        self.minimizer.set(from_file['minimizer'])
        self.set_fit_routine()

        self.norm_with_param.set(from_file['norm_with_param'])
        self.draw_standardized_res.set(from_file['draw_standardized_res'])
        self.use_nbm.set(from_file['use_nbm'])
        self.draw_ppm.set(from_file['draw_ppm'])
        self.draw_rel_peak0.set(from_file['draw_rel_peak0'])
        self.thermo_channel.set(from_file['thermo_channel'])
        self.units = from_file['units']
        self.label_default.set(from_file['label_default'])
        self.ppm_reference = from_file['ppm_reference']
        self.update_period = from_file['update_period']
        self.bnmr_data_dir = from_file['bnmr_data_dir']
        self.bnqr_data_dir = from_file['bnqr_data_dir']

        # set deadtime correction
        self.deadtime = from_file['deadtime']
        self.deadtime_switch.set(from_file['deadtime_switch'])
        self.deadtime_global.set(from_file['deadtime_global'])

        # fileviewer
        fileviewer_tab = self.fileviewer
        fileviewer_tab.year.set(from_file['fileview_year'])
        fileviewer_tab.runn.set(from_file['fileview_run'])

        try:
            fileviewer_tab.get_data()
        except RuntimeError:
            pass
        else:
            fileviewer_tab.asym_type.set(from_file['fileview_asym_type'])
            fileviewer_tab.rebin.set(from_file['fileview_rebin'])
            fileviewer_tab.is_updating.set(from_file['fileview_is_updating'])

        # fetch files
        fetch_files = self.fetch_files
        fetch_files.year.set(from_file['fetch_year'])
        fetch_files.run.set(from_file['fetch_run'])
        fetch_files.check_state.set(from_file['fetch_check_state'])
        fetch_files.check_state_data.set(from_file['fetch_check_state_data'])
        fetch_files.check_state_fit.set(from_file['fetch_check_state_fit'])
        fetch_files.check_state_res.set(from_file['fetch_check_state_res'])
        fetch_files.check_rebin.set(from_file['fetch_check_rebin'])
        fetch_files.check_bin_remove.set(from_file['fetch_check_bin_remove'])
        fetch_files.asym_type.set(from_file['fetch_asym_type'])
        fetch_files.omit_state.set(from_file['fetch_omit_state'])
        fetch_files.base_bins.set(from_file['fetch_base_bins'])

        # load selected runs
        data = from_file['data']
        setyear = fetch_tab.year.get()
        setrun =  fetch_tab.run.get()
        for id, d in data.items():

            # set year and run and fetch
            fetch_tab.year.set(d['year'])
            fetch_tab.run.set(d['run'])
            fetch_tab.get_data()

            # set corresponding parameters for the run
            d_actual = self.data[id]
            d_actual.omit.set(d['omit'])
            d_actual.omit_scan.set(d['omit_scan'])
            d_actual.check_draw_data.set(d['check_draw_data'])
            d_actual.check_draw_fit.set(d['check_draw_fit'])
            d_actual.check_draw_res.set(d['check_draw_res'])
            d_actual.check_state.set(d['check_state'])
            d_actual.label.set(d['label'])
            d_actual.rebin.set(d['rebin'])
            d_actual.base_bins.set(d['base_bins'])

        # reset year and run input info
        fetch_tab.year.set(setyear)
        fetch_tab.run.set(setrun)

        # fit files
        fit_files = self.fit_files
        fit_files.annotation.set(from_file['fit_annotation'])
        fit_files.asym_type.set(from_file['fit_asym_type'])
        fit_files.fit_function_title.set(from_file['fit_fit_function_title'])
        fit_files.n_component.set(from_file['fit_n_component'])
        fit_files.par_label.set(from_file['fit_par_label'])
        fit_files.set_as_group.set(from_file['fit_set_as_group'])
        fit_files.set_prior_p0.set(from_file['fit_set_prior_p0'])
        fit_files.use_rebin.set(from_file['fit_use_rebin'])
        fit_files.xaxis.set(from_file['fit_xaxis'])
        fit_files.yaxis.set(from_file['fit_yaxis'])
        fit_files.xlo.set(from_file['fit_xlo'])
        fit_files.xhi.set(from_file['fit_xhi'])
        fit_files.gchi_label['text'] = from_file['gchi']

        # get parameters in fitting page
        fit_files.populate()
        fit_files.populate_param(force_modify=True)

        # constrain
        constr = fit_files.pop_fitconstr
        constr.constraints_are_set = from_file['constr_constraints_are_set']

        constr.show()
        constr.entry.insert('1.0', from_file['constr_input_fn_text'])
        constr.get_input()

        if constr.constraints_are_set:
            constr.set_constraints()
        else:
            constr.win.destroy()
            constr.win.update()

        # set parameter values
        data = self.data
        fitpar_all = from_file['fitpar']
        for id, fitpar in fitpar_all.items():

            # make sure dataline checkboxes are active
            fetch_tab.data_lines[id].draw_fit_checkbox['state'] = 'normal'
            fetch_tab.data_lines[id].draw_res_checkbox['state'] = 'normal'

            # get pulse length
            pulse_len = 0
            if '2' in data[id].mode:
                pulse_len = data[id].pulse_s

            # get probe lifetime
            lifetime = bd.life[from_file['probe_species']]

            # get fit function
            fitfn = fit_files.fitter.get_fn(from_file['fit_fit_function_title'],
                                            from_file['fit_n_component'],
                                            pulse_len,
                                            lifetime)

            if '2' in data[id].mode and from_file['probe_species'] == 'Mg31':
                fitfn1 = decay_corrected_fn(fa_31Mg, fitfn, pulse_len)
            else:
                fitfn1 = fitfn

            # set fit results
            fitpar = pd.DataFrame(fitpar)
            data[id].set_fitpar(fitpar)
            data[id].drop_unused_param(fitpar.index)
            data[id].fitfn = fitfn1

        fit_files.populate()

    # ======================================================================= #
    def on_closing(self):
        """Excecute this when window is closed: destroy and close all plots."""
        self.logger.info('Closing all windows.')
        plt.close('all')
        self.root.destroy()
        self.logger.info('Finished     ' + '-'*50)

    # ======================================================================= #
    def report_issue(self):
        """Display github issue page"""
        self.logger.info('Opening github issue page: '+\
                         'https://github.com/dfujim/bfit/issues')
        webbrowser.open('https://github.com/dfujim/bfit/issues')

    # ======================================================================= #
    def return_binder(self, *args):
        """Switch between various functions of the enter button. """

        idx = self.notebook.index('current')
        self.logger.debug('Calling return key command for notebook index %d', idx)
        if idx == 0:        # data viewer
            self.fileviewer.get_data()
        elif idx == 1:        # data fetch_files
            self.fetch_files.return_binder()
        elif idx == 2:        # fit viewer
            self.fit_files.return_binder()
        else:
            pass

    # ======================================================================= #
    def save_state(self, filename=None):
        """
            Save the state of the gui:

            dataline state info
            Fitting function
            Number of components
            Initial inputs
            Fit results
        """

        # final output
        to_file = {}

        # bfit menu options
        to_file['probe_species'] = self.probe_species.get()
        to_file['minimizer'] = self.minimizer.get()
        to_file['norm_with_param'] = self.norm_with_param.get()
        to_file['draw_standardized_res'] = self.draw_standardized_res.get()
        to_file['use_nbm'] = self.use_nbm.get()
        to_file['draw_ppm'] = self.draw_ppm.get()
        to_file['draw_rel_peak0'] = self.draw_rel_peak0.get()
        to_file['thermo_channel'] = self.thermo_channel.get()
        to_file['units'] = self.units
        to_file['label_default'] = self.label_default.get()
        to_file['ppm_reference'] = self.ppm_reference
        to_file['update_period'] = self.update_period
        to_file['deadtime'] = self.deadtime
        to_file['deadtime_switch'] = self.deadtime_switch.get()
        to_file['deadtime_global'] = self.deadtime_global.get()
        to_file['style'] = self.style
        to_file['hist_select'] = self.hist_select
        to_file['draw_style'] = self.draw_style.get()
        to_file['draw_fit'] = self.draw_fit.get()
        to_file['bnmr_data_dir'] = self.bnmr_data_dir
        to_file['bnqr_data_dir'] = self.bnqr_data_dir

        # fileviewer
        fileviewer_tab = self.fileviewer
        to_file['fileview_year'] = fileviewer_tab.year.get()
        to_file['fileview_run'] = fileviewer_tab.runn.get()
        to_file['fileview_asym_type'] = fileviewer_tab.asym_type.get()
        to_file['fileview_rebin'] = fileviewer_tab.rebin.get()
        to_file['fileview_is_updating'] = fileviewer_tab.is_updating.get()

        # fetch files
        fetch_files = self.fetch_files
        to_file['fetch_year'] = fetch_files.year.get()
        to_file['fetch_run'] = fetch_files.run.get()
        to_file['fetch_check_state'] = fetch_files.check_state.get()
        to_file['fetch_check_state_data'] = fetch_files.check_state_data.get()
        to_file['fetch_check_state_fit'] = fetch_files.check_state_fit.get()
        to_file['fetch_check_state_res'] = fetch_files.check_state_res.get()
        to_file['fetch_check_rebin'] = fetch_files.check_rebin.get()
        to_file['fetch_check_bin_remove'] = fetch_files.check_bin_remove.get()
        to_file['fetch_asym_type'] = fetch_files.asym_type.get()
        to_file['fetch_base_bins'] = fetch_files.base_bins.get()
        to_file['fetch_omit_state'] = fetch_files.omit_state.get()

        # get state from data
        datalines = fetch_files.data_lines
        data = {}
        for id, d in self.data.items():
            data[id] = {
                    'omit'              :d.omit.get(),
                    'check_draw_data'   :d.check_draw_data.get(),
                    'check_draw_fit'    :d.check_draw_fit.get(),
                    'check_draw_res'    :d.check_draw_res.get(),
                    'check_state'       :d.check_state.get(),
                    'omit_scan'         :d.omit_scan.get(),
                    'id'                :d.id,
                    'label'             :d.label.get(),
                    'rebin'             :d.rebin.get(),
                    'run'               :d.run,
                    'year'              :d.year,
                    'base_bins'         :d.base_bins.get(),
                    }
        to_file['data'] = data

        # fit files
        fit_files = self.fit_files
        to_file['gchi'] = fit_files.gchi_label['text']
        to_file['fit_annotation'] = fit_files.annotation.get()
        to_file['fit_asym_type'] = fit_files.asym_type.get()
        to_file['fit_fit_function_title'] = fit_files.fit_function_title.get()
        to_file['fit_n_component'] = fit_files.n_component.get()
        to_file['fit_par_label'] = fit_files.par_label.get()
        to_file['fit_set_as_group'] = fit_files.set_as_group.get()
        to_file['fit_set_prior_p0'] = fit_files.set_prior_p0.get()
        to_file['fit_use_rebin'] = fit_files.use_rebin.get()
        to_file['fit_xaxis'] = fit_files.xaxis.get()
        to_file['fit_yaxis'] = fit_files.yaxis.get()
        to_file['fit_xlo'] = fit_files.xlo.get()
        to_file['fit_xhi'] = fit_files.xhi.get()

        # save constraints
        constr = fit_files.pop_fitconstr
        to_file['constr_constraints_are_set'] = constr.constraints_are_set
        to_file['constr_input_fn_text'] = constr.input_fn_text

        # get parameter values from data
        data_values = {id: dat.fitpar.to_dict() for id, dat in self.data.items()}
        to_file['fitpar'] = data_values

        # get xlims
        to_file['xlo'] = fit_files.xlo.get()
        to_file['xhi'] = fit_files.xhi.get()

        # save file ----------------------------------------------------------
        if filename is None:
            fid = filedialog.asksaveasfile(mode='w', filetypes=[('yaml', '*.yaml'),
                                                           ('allfiles', '*')],
                                           defaultextension='.yaml')
        else:
            fid = open(filename, 'w')

        if fid:
            yaml.dump(to_file, fid)
            fid.close()

        self.logger.info('Saving program state to %s', fid)

    # ======================================================================= #
    def scroll_binder(self, event):
        """
            Switch between various functions of the mousewheel button.
            Bound to <Button-4> and <Button-5>
        """

        idx = self.notebook.index('current')

        if idx == 0:        # data viewer
            pass
        elif idx == 1:        # data fetch_files
            self.fetch_files.canvas_scroll(event)
        elif idx == 2:        # fit viewer
            self.fit_files.canvas_scroll(event)
        else:
            pass

    # ======================================================================= #
    def search_archive(self):
        self.logger.info('Opening mud archive musr website')
        webbrowser.open('http://musr.ca/mud/runSel.html', new=1)

    # ======================================================================= #
    def set_bnmr_dir(self):
        """Set directory location via environment variable BNMR_ARCHIVE."""
        self.logger.info('Setting BNMR environment directory')
        d = filedialog.askdirectory(parent=self.root, mustexist=True,
                initialdir=self.bnmr_data_dir)

        if type(d) == str:
            self.bnmr_data_dir = d
            os.environ[self.bnmr_archive_label] = d
            self.logger.debug('Environment variable "%s" set to "%s"',
                              self.bnmr_archive_label, d)
        else:
            self.logger.error('Input was not of type string')

    # ======================================================================= #
    def set_bnqr_dir(self):
        """Set directory location via environment variable BNQR_ARCHIVE."""
        self.logger.info('Setting BNQR environment directory')
        d = filedialog.askdirectory(parent=self.root, mustexist=True,
                initialdir=self.bnqr_data_dir)

        if type(d) == str:
            self.bnqr_data_dir = d
            os.environ[self.bnqr_archive_label] = d
            self.logger.debug('Environment variable "%s" set to "%s"',
                              self.bnqr_archive_label, d)
        else:
            self.logger.error('Input was not of type string')

    # ======================================================================= #
    def set_fit_routine(self):
        """Set python module for fitting routines"""

        self.logger.info('Loading module...')
        self.routine_mod = importlib.import_module(self.minimizer.get())

        # repopuate fitter
        self.logger.info('Repopulating fitter...')
        self.fit_files.fitter = self.routine_mod.fitter(
                                    keyfn = self.get_run_key,
                                    probe_species = self.probe_species.get())
        self.fit_files.fit_routine_label['text'] = self.fit_files.fitter.__name__
        self.fit_files.populate()
        self.logger.debug('Success.')

    # ======================================================================= #
    def set_fit_routine_with_popup(self):
        """Set python module for fitting routines"""
        self.logger.info('Setting fitting backend routine')
        d = filedialog.askopenfilename(initialdir = "./",
                title = "Select fitting routine module",
                filetypes = (("python modules", "*.py"),
                             ("cython modules", "*.pyx"),
                             ("all files", "*.*")))

        if type(d) == str:

            # empty condition
            if d == '':
                self.logger.error('Input was empty string.')
                return

            # get paths
            path = os.path.abspath(d)
            pwd = os.getcwd()

            # load the module
            os.chdir(os.path.dirname(path))
            self.minimizer.set(os.path.splitext(os.path.basename(d))[0])
            self.set_fit_routine()
            os.chdir(pwd)

        else:
            self.logger.error('Input was not of type string.')

    # ======================================================================= #
    def set_icon(self, window):
        """Set the icon for new windows"""
        try:
            img = PhotoImage(file=icon_path)
            window.tk.call('wm', 'iconphoto', window._w, img)
        except Exception as err:
            print(err)

    # ======================================================================= #
    def set_matplotlib(self):
        """Edit matplotlib settings file, or give info on how to do so."""

        self.logger.info('Attempting to edit matplotlibrc file')

        # settings
        location = os.path.join(os.environ['HOME'], '.config', 'matplotlib')
        filename = "matplotlibrc"
        weblink = 'http://matplotlib.org/users/customizing.html'+\
                  '#the-matplotlibrc-file'

        # check for file existance
        if not os.path.isfile(os.path.join(location, filename)):
            self.logger.debug('File not found.')
            value = messagebox.showinfo(parent=self.mainframe,
                    title="Get matplotlibrc", \
                    message="No matplotlibrc file found.",
                    detail="Press ok to see web resource.",
                    type='okcancel')

            if value == 'ok':
                webbrowser.open(weblink)
            return

        # if file exists, edit
        self.logger.debug('File found. Opening in external program.')
        subprocess.call(['xdg-open', os.path.join(location, filename)])

    # ======================================================================= #
    def set_1f_shift_style(self, clicked_style):
        """
            Ensure that ppm and peak shift modes are not both selected at the
            same time.

            clicked_style: string, one of 'ppm' or 'peak'
        """

        if clicked_style == 'ppm':
            self.draw_rel_peak0.set(False)

        elif clicked_style == 'peak':
            self.draw_ppm.set(False)

    # ======================================================================= #
    def set_all_labels(self, *a):    self.fetch_files.set_all_labels()
    def set_check_all(self, x):
        self.logger.info('Checking all files')
        state = self.fetch_files.check_state.get()
        self.fetch_files.check_state.set(not state)
        self.fetch_files.check_all()
    def set_deadtime(self):          popup_deadtime(wref.proxy(self))
    def set_draw_style(self):        popup_drawstyle(wref.proxy(self))
    def set_histograms(self, *a):    popup_set_histograms(wref.proxy(self))
    def set_focus_tab(self, idn, *a): self.notebook.select(idn)
    def set_nbm(self, mode):
        """
            Set the nbm variable based on the run mode
        """

        # get new nbm BooleanVar
        new_nbm = self.nbm_dict.get(mode, self.nbm_dict[''])

        # switch the variable
        menu = self.menus['Draw Mode']
        try:
            idx = menu.index('Use NBM in asymmetry')
        except TclError:
            idx = menu.index('Use NBM in asymmetry (1n)')

        menu.entryconfig(idx, variable = new_nbm)
        self.use_nbm = new_nbm

        # change the menu label
        if mode == '1n':
            menu.entryconfig(idx, label = 'Use NBM in asymmetry (1n)')
        else:
            menu.entryconfig(idx, label = 'Use NBM in asymmetry')
    def set_ppm_reference(self, *a): popup_set_ppm_reference(wref.proxy(self))
    def set_probe_species(self, *a):
        species = self.probe_species.get()
        self.fit_files.fitter.probe_species = species
        self.fit_files.probe_label['text'] = species
        self.logger.info('Probe species changed to %s', species)
    def set_redraw_period(self, *a): popup_redraw_period(wref.proxy(self))
    def set_units(self, *a):         popup_units(wref.proxy(self))
    def set_style_new(self, x):
        self.logger.info('Setting draw style to "new"')
        self.draw_style.set('new')
    def set_style_stack(self, x):
        self.logger.info('Setting draw style to "stack"')
        self.draw_style.set('stack')
    def set_style_redraw(self, x):
        self.logger.info('Setting draw style to "redraw"')
        self.draw_style.set('redraw')
    def set_tab_change(self, event):

        new_tab = event.widget.index("current")

        self.logger.debug('Changing to tab %d', new_tab)

        # fileviewer
        if new_tab == 0:
            self.fileviewer.set_nbm()

        # fetch files
        elif new_tab == 1:
            self.fetch_files.set_nbm()

        # fit files
        elif new_tab == 2:
            self.fit_files.populate()

    def set_thermo_channel(self, ):
        self.fetch_files.update_data()
        self.fileviewer.get_data()

    # ======================================================================= #
    def set_asym_calc_mode_box(self, mode, parent, area, *args):
        """Set asym combobox values. Asymmetry calculation and draw modes."""

        self.logger.debug('Setting asym combobox values for mode '+\
                         '"%s" and area "%s"', mode, area)

        # get list of possible run modes
        try:
            modes = list(self.asym_dict_keys[mode])
        except KeyError as err:
            messagebox.showerror('Error', 'Unknown mode %s' % mode)
            self.logger.exception(str(err))
            raise err from None

        # prune the list to match only ok files
        if parent == self.fit_files:
            modes = [m for m in modes if self.asym_dict[m] in \
                                        self.fit_files.fitter.valid_asym_modes]

        # add single counter asymetries based on mode and type
        if mode != '2e':
            if 'BNMR' in area:
                modes.extend(['Forward Counter', 'Backward Counter'])
            if 'BNQR' in area:
                modes.extend(['Left Counter', 'Right Counter'])

        # selection: switch if run mode not possible
        if parent.asym_type.get() not in modes:
            parent.asym_type.set(modes[0])

        # set list
        parent.entry_asym_type['values'] = modes

    # ======================================================================= #
    def update_bfit(self):
        """Check pip for updated version"""

        self.logger.info('Using pip to update')

        # set up queue to get signal
        que = Queue()

        # run update function
        def do_update():
            try:
                subprocess.call([sys.executable, "-m", "pip", "install",
                                 "--upgrade", 'bfit'])
            except subprocess.CalledProcessError:
                que.put('Error')
            else:
                que.put('Success')

        popup = popup_ongoing_process(self,
                            target = do_update,
                            message="Updating bfit...",
                            queue = que)
        output = popup.run()

        # end update
        if output == 'Error':
            messagebox.showerror("Error", 'Update error. Use command line to update.')
        elif output == 'Success':
            messagebox.showinfo("Success", 'Close window and restart bfit to implement updates.')
        else:
            return

    # ======================================================================= #
    def whatsnew(self):
        """Display releases page"""
        self.logger.info('Opening https://github.com/dfujim/bfit/releases')
        webbrowser.open('https://github.com/dfujim/bfit/releases')

# =========================================================================== #
if __name__ == "__main__":
    bfit()




# Generate initial parameters for some common functions
# Derek Fujimoto
# Oct 2021

import numpy as np
import pandas as pd

# ======================================================================= #
def gen_init_par(fn_name, ncomp, bdataobj, asym_mode='combined'):
    """Generate initial parameters for a given function.

        fname: name of function. Should be the same as the param_names keys
        ncomp: number of components
        bdataobj: a bdata object representative of the fitting group.
        asym_mode: what kind of asymmetry to fit

        Set and return pd.DataFrame of initial parameters.
            col: p0, blo, bhi, fixed
            index: parameter name
    """

    fn_name = fn_name.lower()

    # asym_mode un-used types
    if asym_mode in (   'h',           # Split Helicity
                        'hm',          # Matched Helicity
                        'hs',          # Shifted Split
                        'cs',          # Shifted Combined
                        'hp',          # Matched Peak Finding
                        'r',           # Raw Scans
                        'rhist',       # Raw Histograms
                        '2e_raw_c',    # Combined Hel Raw
                        '2e_raw_h',    # Split Hel Raw
                        '2e_sl_h',     # Split Hel Slopes
                        '2e_dif_h',    # Split Hel Diff
                        'ad',          # Alpha Diffusion
                        "at_c",        # Combined Hel (Alpha Tag)
                        "at_h",        # Split Hel (Alpha Tag)
                        "nat_c",       # Combined Hel (!Alpha Tag)
                        "nat_h",       # Split Hel (!Alpha Tag)
                    ):
        errmsg = "Asymmetry calculation type not implemented for fitting"
        raise RuntimeError(errmsg)

    # get asymmetry
    x, a, da = bdataobj.asym(asym_mode)

    # set pulsed exp fit initial parameters
    if fn_name in ('exp', 'bi exp', 'str exp', 'biexp', 'strexp'):
        # ampltitude average of first 5 bins
        amp = abs(np.mean(a[0:5])/ncomp)

        # T1: time after beam off to reach 1/e
        idx = int(bdataobj.ppg.beam_on.mean)
        beam_duration = x[idx]
        amp_beamoff = a[idx]
        target = amp_beamoff/np.exp(1)

        x_target = x[np.sum(a>target)]
        T1 = abs(x_target-beam_duration)

        # bounds and amp
        if asym_mode == 'n':
            amp_bounds = (-np.inf, np.inf)
            amp = -amp
        else:
            amp_bounds = (0, np.inf)

        # set values
        if fn_name == 'exp':
            par_values = {  '1_T1':(1./T1, 0, np.inf, False),
                            'amp':(amp, *amp_bounds, False),
                         }

        elif fn_name in ('bi exp', 'biexp'):
            par_values = {  '1_T1':(1./T1, 0, np.inf, False),
                            '1_T1b':(10./T1, 0, np.inf, False),
                            'fraction_b':(0.5, 0, 1, False),
                            'amp':(amp, *amp_bounds, False),
                         }

        elif fn_name in ('str exp', 'strexp'):
            par_values = {  '1_T1':(1./T1, 0, np.inf, False),
                            'beta':(0.5, 0, 1, False),
                            'amp':(amp, *amp_bounds, False),
                         }

    # set time integrated fit initial parameters
    elif fn_name in ('lorentzian', 'gaussian', 'bilorentzian', 'quadlorentz', 'pseudovoigt'):

        # get baseline
        base = np.mean(a[:5])

        # check for upside down helicities (peak going up)
        avg = np.mean(a)
        if base < avg:
            amin = max(a[a!=0])
        else:
            amin = min(a[a!=0])

        # get peak asym value
        peak = x[np.where(a==amin)[0][0]]
        height = base-amin
        width = 2*abs(peak-x[np.where(a<amin+height/2)[0][0]])

        # bounds
        if asym_mode == 'n':
            height_bounds = [-np.inf, 0]
        else:
            height_bounds = [0, np.inf]

        # check bounds validity
        if height < height_bounds[0]:
            height_bounds[0] = -np.inf

        if height > height_bounds[1]:
            height_bounds[1] = np.inf

        # set values (value, low bnd, high bnd, fixed)
        if fn_name == 'lorentzian':
            par_values = {'peak':(peak, min(x), max(x), False),
                          'fwhm':(width, 0, np.inf, False),
                          'height':(height, *height_bounds, False),
                          'baseline':(base, -np.inf, np.inf, False)
                         }
        elif fn_name == 'gaussian':
            par_values = {'mean':(peak, min(x), max(x), False),
                          'sigma':(width, 0, np.inf, False),
                          'height':(height, *height_bounds, False),
                          'baseline':(base, -np.inf, np.inf, False)
                          }
        elif fn_name == 'bilorentzian':
            par_values = {'peak':(peak, min(x), max(x), False),
                          'fwhmA':(width*10, 0, np.inf, False),
                          'heightA':(height/10, *height_bounds, False),
                          'fwhmB':(width, 0, np.inf, False),
                          'heightB':(height*9/10, *height_bounds, False),
                          'baseline':(base, -np.inf, np.inf, False)
                         }
        elif fn_name == 'quadlorentz':

            dx = max(x)-min(x)
            par_values = {'nu_0':((max(x)+min(x))/2, min(x), max(x), False),
                          'nu_q':(dx/12, 0, dx, False),
                          'efgAsym':(0, 0, 1, True),
                          'efgTheta':(0, 0, 2*np.pi, True),
                          'efgPhi':(0, 0, 2*np.pi, True),
                          'amp0':(height, height*0.1, np.inf, False),
                          'amp1':(height, height*0.1, np.inf, False),
                          'amp2':(height, height*0.1, np.inf, False),
                          'amp3':(height, height*0.1, np.inf, False),
                          'fwhm':(dx/10, 0, dx, False),
                          'baseline':(base, -np.inf, np.inf, False)
                         }
        elif fn_name == 'pseudovoigt':

            par_values = {'peak':(peak, min(x), max(x), False),
                          'fwhm':(width, 0, np.inf, False),
                          'height':(height, *height_bounds, False),
                          'fracL':(0.5, 0, 1, False),
                          'baseline':(base, -np.inf, np.inf, False)
                         }

    else:
        raise RuntimeError(f'gen_init_par: Bad function name "{fn_name}".')

    # do multicomponent
    par_values2 = {}
    if ncomp > 1:
        for c in range(ncomp):
            for n in par_values.keys():
                if 'baseline' not in n:
                    par_values2[n+'_%d' % c] = par_values[n]
                else:
                    par_values2[n] = par_values[n]
    else:
        par_values2 = par_values

    return pd.DataFrame(par_values2, index=['p0', 'blo', 'bhi', 'fixed']).transpose()




# File viewer tab for bfit
# Derek Fujimoto
# Nov 2017

from tkinter import *
from tkinter import ttk
from tkinter import filedialog
from multiprocessing import Process, Pipe
from bdata.calc import nqr_B0_hh6, nqr_B0_hh3
from bfit.backend.fitdata import fitdata
from bdata import bdata, bmerged
from bfit import logger_name
import bdata as bd

import numpy as np
import matplotlib.pyplot as plt
import sys, os, time, glob, datetime
import logging


__doc__ = """
    View file contents tab.
    
    To-do:
        cumulative count viewing
    """

# =========================================================================== #
class fileviewer(object):
    """
        Data fields:
            asym_type: StringVar, drawing style
            bfit: bfit object
            button_run_from_file: ttk.Button for load from run
            data: bdata object for drawing
            entry_asym_type: combobox for asym calculations
            entry_rebin: Spinbox
            entry_runn: Spinbox
            entry_year: Spinbox
            filename: string, path to file to load, disabled if == ''
            is_updating: BooleanVar, True if update draw
            rebin: IntVar() rebin value
            runn: IntVar() run number
            text: Text widget for displaying run information
            update_id: string, run id for the currently updating run
            year: IntVar() year of exp 
    """
    
    default_export_filename = "%d_%d.csv" # year_run.csv
    update_id = ''
    mode_dict = {"1c":"Camp Scan", 
                 "1d":"Laser Scan", 
                 "1e":"Field Scan", 
                 "1w":"Frequency Comb", 
                 "1f":"Frequency Scan", 
                 "1x":"Frequency Scan", 
                 "1n":"Rb Cell Scan", 
                 "20":"SLR", 
                 '2e':'Randomized Frequency Scan',
                 '2h':'SLR with Alpha Tracking', 
                 '2s':'Spin Echo', 
                }
    
    mode_epics_var = {'1n':'ILE2:BIAS15:VOL',
                      '1d':'ILE2:RING:FREQ:VOL',
                      '1e':'ILE2A1:HH:CUR',
                     }
    
    # ======================================================================= #
    def __init__(self, file_tab, bfit):
        """ Position tab tkinter elements"""
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing')
        
        # year and filenumber entry ------------------------------------------
        entry_frame = ttk.Frame(file_tab, borderwidth=1)
        self.year = IntVar()
        self.runn = IntVar()
        self.rebin = IntVar()
        self.filename = ''
        self.bfit = bfit
        
        self.year.set(self.bfit.get_latest_year())
        self.rebin.set(1)
        
        self.entry_year = Spinbox(entry_frame, \
                from_=2000, to=np.inf, 
                textvariable=self.year, width=5)
        self.entry_runn = Spinbox(entry_frame, \
                from_=0, to=np.inf, 
                textvariable=self.runn, width=7)
        self.runn.set(40000)
        
        # fetch button
        fetch = ttk.Button(entry_frame, text='Fetch', command=self.get_data)
            
        # draw button
        draw = ttk.Button(entry_frame, text='Draw', 
                          command=lambda:self.draw(figstyle='inspect'))
        
        # get run from file
        self.button_run_from_file = ttk.Button(file_tab, text='Load run from file', 
                                                command=self.load_file)
        
        # grid and labels
        entry_frame.grid(column=0, row=0, sticky=N, columnspan=2)
        ttk.Label(entry_frame, text="Year:").grid(column=0, row=0, sticky=E)
        self.entry_year.grid(column=1, row=0, sticky=E)
        ttk.Label(entry_frame, text="Run Number:").grid(column=2, row=0, sticky=E)
        self.entry_runn.grid(column=3, row=0, sticky=E)
        fetch.grid(column=4, row=0, sticky=E)
        draw.grid(column=5, row=0, sticky=E)
        self.button_run_from_file.grid(column=1, row=0, sticky=E, padx=5)
        
        # padding 
        for child in entry_frame.winfo_children(): 
            child.grid_configure(padx=5, pady=5)

        # viewer frame -------------------------------------------------------
        view_frame = ttk.Frame(file_tab, borderwidth=2)
        
        self.text_nw = Text(view_frame, width=88, height=20, state='normal', wrap='none')
        self.text_ne = Text(view_frame, width=88, height=20, state='normal', wrap='none')
        self.text_sw = Text(view_frame, width=88, height=20, state='normal', wrap='none')
        self.text_se = Text(view_frame, width=88, height=20, state='normal', wrap='none')
        
        ttk.Label(view_frame, text="Run Info").grid(column=0, row=0, sticky=N, pady=5)
        ttk.Label(view_frame, text="PPG Parameters").grid(column=1, row=0, sticky=N, pady=5)
        ttk.Label(view_frame, text="Camp and Stats").grid(column=0, row=2, sticky=N, pady=5)
        ttk.Label(view_frame, text="EPICS").grid(column=1, row=2, sticky=N, pady=5)
        
        self.text_nw.grid(column=0, row=1, sticky=(N, W, E, S), padx=5)
        self.text_ne.grid(column=1, row=1, sticky=(N, W, E, S), padx=5)
        self.text_sw.grid(column=0, row=3, sticky=(N, W, E, S), padx=5)
        self.text_se.grid(column=1, row=3, sticky=(N, W, E, S), padx=5)
        
        view_frame.grid(column=0, row=1, sticky=(N, E, W), columnspan=2)
        
        # details frame: stuff at the bottom ----------------------------------
        details_frame = ttk.Frame(file_tab)
        self.entry_rebin = Spinbox(details_frame, from_=1, to=100, width=3, \
                textvariable=self.rebin)
        
        # update check box
        self.is_updating = BooleanVar()
        self.is_updating.set(False)
        update_box = ttk.Checkbutton(details_frame, text='Periodic Redraw', 
                command=self.do_update, variable=self.is_updating, onvalue=True, 
                offvalue=False)

        # asymmetry type combobox
        self.asym_type = StringVar()
        self.asym_type.set('')
        self.entry_asym_type = ttk.Combobox(details_frame, \
                textvariable=self.asym_type, state='readonly', width=25)
        self.entry_asym_type['values'] = ()
                
        # gridding
        ttk.Label(details_frame, text="Rebin:").grid(column=0, row=0, sticky=E)
        self.entry_rebin.grid(column=1, row=0, sticky=E)
        self.entry_asym_type.grid(column=2, row=0, sticky=E)
        update_box.grid(column=3, row=0, sticky=E)
        details_frame.grid(column=0, row=2, sticky=S, columnspan=2)
        
        # padding 
        for child in details_frame.winfo_children(): 
            child.grid_configure(padx=5, pady=5)
            
        # resizing
        file_tab.grid_rowconfigure(1, weight=1)
        file_tab.grid_columnconfigure(0, weight=1)
        
        entry_frame.grid_columnconfigure(0, weight=2)
        entry_frame.grid_columnconfigure(2, weight=1)
        entry_frame.grid_rowconfigure(0, weight=1)
        
        for i in range(2):
            view_frame.grid_columnconfigure(i, weight=1)
        view_frame.grid_rowconfigure(1, weight=1)
        view_frame.grid_rowconfigure(3, weight=1)
        
        for t in [self.text_nw, self.text_ne, self.text_sw, self.text_se]:
            for i in range(5):
                t.grid_columnconfigure(i, weight=1)
                t.grid_rowconfigure(i, weight=1)
            
        self.logger.debug('Initialization success.')
            
    # ======================================================================= #
    def __del__(self):
        pass
        
    # ======================================================================= #
    def _get_data(self, bdata_obj, quiet=False):
        """
            Display data and send bdata object to bfit draw list. 
            Return True on success, false on Failure
        """
        
        # get data
        try: 
            data = fitdata(self.bfit, bdata_obj)
        except ValueError as err:
            self.logger.exception('File read failed.')
            self.logger.debug(err)
            raise err from None
            for t in [self.text_nw, self.text_sw, self.text_se, self.text_ne]:
                self.set_textbox_text(t, 'File read failed.')
            return False
        except RuntimeError as err:
            self.logger.exception('File does not exist.')
            self.logger.debug(err)
            for t in [self.text_nw, self.text_sw, self.text_se, self.text_ne]:
                self.set_textbox_text(t, 'File does not exist.')
            return False
        
        # set data field
        self.data = data
        
        # set draw parameters
        self.bfit.set_asym_calc_mode_box(data.mode, self, data.area)
        
        # set nbm variable
        self.set_nbm()
        
        # set rebin
        self.set_rebin(self.data)
        
        # quiet mode: don't update text
        if quiet: return True
        
        # NW -----------------------------------------------------------------
        
        # get data: headers
        mode = self.mode_dict[data.mode]
        try:
            if data.ppg.rf_enable.mean and data.mode == '20' and \
                                                        data.ppg.rf_on.mean > 0:
                mode = "Hole Burning"
        except AttributeError:
            pass
        
        mins, sec = divmod(data.duration, 60)
        duration = "%dm %ds" % (mins, sec)
        
        # set dictionary
        data_nw =  {"Run":'%d (%d)' % (data.run, data.year), 
                    "Area": data.area, 
                    "Run Mode": "%s (%s)" % (mode, data.mode), 
                    "Title": data.title, 
                    "Experimenters": data.experimenter, 
                    "Sample": data.sample, 
                    "Orientation":data.orientation, 
                    "Experiment":str(data.exp), 
                    "Run Duration": duration, 
                    "Start": data.start_date, 
                    "End": data.end_date, 
                    "":"", 
                    }
        
        # set key order 
        key_order_nw = ['Run', 'Run Mode', 'Title', '', 
                        'Start', 'End', 'Run Duration', '', 
                        'Sample', 'Orientation', '', 
                        'Experiment', 'Area', 'Experimenters','',
                        ]
        
        # get probe species
        try:
            probe = data.ppg.probe_species.units
            data_nw["Probe"] = str(probe)
            key_order_nw.append('Probe')
            key_order_nw.append('')
        except AttributeError:
            pass
        
        # comments
        if hasattr(data, 'comments'):
            data_nw["Comment"] = data.comments['Comments during run'].body
            key_order_nw.append('Comment')
        
        # SW -----------------------------------------------------------------
        data_sw = {'':''}
        key_order_sw = []
                        
        # get data: temperature and fields
        try:
            temp = data.temperature.mean
            temp_stdv = data.temperature.std
            data_sw["Temperature"] = "%.2f +/- %.2f K" % (temp, temp_stdv)
            key_order_sw.append('Temperature')
        except AttributeError:
            pass
        
        try:
            curr = data.camp.smpl_current
            data_sw["Heater Current"] = "%.2f +/- %.2f A" % (curr.mean, curr.std)
            key_order_sw.append('Heater Current')
        except AttributeError:
            pass
        
        try:
            temp = data.camp.oven_readC.mean
            temp_stdv = data.camp.oven_readC.std
            data_sw['Oven Temperature'] = "%.2f +/- %.2f K" % (temp, temp_stdv)
            key_order_sw.append('Oven Temperature')
        except AttributeError:
            pass
        
        try:
            curr = data.camp.oven_current
            data_sw['Oven Current'] = "%.2f +/- %.2f A" % (curr.mean, curr.std)
            key_order_sw.append('Oven Current')
        except AttributeError:
            pass
        
        try: 
            field = np.around(data.camp.b_field.mean, 3)
            field_stdv = np.around(data.camp.b_field.std, 3)
            data_sw['Magnetic Field'] = "%.3f +/- %.3f T" % (field, field_stdv)
            key_order_sw.append('Magnetic Field')
        except AttributeError:
            pass
            
        try: 
            val, _ = nqr_B0_hh3(amps=data.epics.hh_current.mean)
            data_sw['Magnetic Field'] = "%.3f Gauss" % val
            key_order_sw.append('Magnetic Field')
        except AttributeError:
            pass
            
        if key_order_sw[-1] != '':
            key_order_sw.append('')
                
        # cryo options
        try: 
            mass = data.camp.mass_read
            data_sw['Mass Flow'] = "%.3f +/- %.3f" % (mass.mean, mass.std)
            key_order_sw.append('Mass Flow')
        except AttributeError:
            pass
    
        try: 
            mass = data.camp.he_read
            data_sw['Mass Flow'] = "%.3f +/- %.3f" % (mass.mean, mass.std)
            key_order_sw.append('Mass Flow')
        except AttributeError:
            pass
    
        try: 
            cryo = data.camp.cryo_read
            data_sw['CryoEx Mass Flow'] = "%.3f +/- %.3f" % (cryo.mean, cryo.std)
            key_order_sw.append('CryoEx Mass Flow')
        except AttributeError:
            pass    
            
        try: 
            data_sw['Needle Setpoint'] = "%.3f turns" % data.camp.needle_set.mean
            key_order_sw.append('Needle Setpoint')
        except AttributeError:
            pass    
            
        try: 
            data_sw['Needle Readback'] = "%.3f turns" % data.camp.needle_pos.mean
            key_order_sw.append('Needle Readback')
        except AttributeError:
            pass    
            
        try:
            lift_set = np.around(data.camp.clift_set.mean, 3)
            data_sw['Cryo Lift Setpoint'] = "%.3f mm" % lift_set
            key_order_sw.append('Cryo Lift Setpoint')
        except AttributeError:
            pass
        
        try:
            lift_read = np.around(data.camp.clift_read.mean, 3)
            data_sw['Cryo Lift Readback'] = "%.3f mm" % lift_read
            key_order_sw.append('Cryo Lift Readback')
        except AttributeError:
            pass
    
        # rf dac
        if mode != 'SLR':
            if key_order_sw[-1] != '':
                key_order_sw.append('')
            
            try: 
                data_sw['rf_dac'] = "%d" % int(data.camp.rf_dac.mean)
                key_order_sw.append('rf_dac')
            except AttributeError:
                pass
            
            try: 
                data_sw['RF Amplifier Gain'] = "%.2f" % data.camp.rfamp_rfgain.mean
                key_order_sw.append('RF Amplifier Gain')
            except AttributeError:
                pass    
        
        if key_order_sw[-1] != '':
            key_order_sw.append('')
        
        # rates and counts
        hist = ('F+', 'F-', 'B-', 'B+') if data.area.upper() == 'BNMR' \
                                     else ('L+', 'L-', 'R-', 'R+')

        try:     
            val = int(np.sum([data.hist[h].data for h in hist]))
            val, unit_val = num_prefix(val)
            data_sw['Total Counts Sample'] = "%.3g%s" % (val, unit_val)
            key_order_sw.append('Total Counts Sample')
        except (AttributeError, KeyError):
            pass
        
        try: 
            val = int(np.sum([data.hist[h].data for h in hist])/data.duration)
            val, unit_val = num_prefix(val)
            data_sw['Total Rate Sample'] =  "%.3g%s (cps)" % (val, unit_val)
            key_order_sw.append('Total Rate Sample')
        except (AttributeError, KeyError):
            pass
        
        try: 
            tag_F = 'F' if data.area == 'BNMR' else 'L'
            tag_B = 'B' if data.area == 'BNMR' else 'R'
    
            F = np.sum([data.hist[h].data for h in hist if tag_F in h])/data.duration
            B = np.sum([data.hist[h].data for h in hist if tag_B in h])/data.duration
            
            F, unit_F = num_prefix(F)
            B, unit_B = num_prefix(B)
            
            try:
                ratio = F/B
            except ZeroDivisionError:
                ratio = np.nan
            
            data_sw['Rate %s/%s' % (tag_F, tag_B)] = \
                    '%.3g%s / %.3g%s (cps) [ratio: %.2f]' % (F, unit_F, B, unit_B, ratio)
            key_order_sw.append('Rate %s/%s' % (tag_F, tag_B))
        except (AttributeError, KeyError):
            pass
        
        if key_order_sw[-1] != '':
            key_order_sw.append('')
        
        hist = ('F+', 'F-', 'B-', 'B+')    
        try: 
            val = int(np.sum([data.hist['NBM'+h].data for h in hist]))
            val, unit_val = num_prefix(val)
            data_sw['Total Counts NBM'] = "%.3g%s" % (val, unit_val)
            key_order_sw.append('Total Counts NBM')
        except (AttributeError, KeyError):
            pass
        
        try: 
            val = int(np.sum([data.hist['NBM'+h].data for h in hist])/data.duration)
            val, unit_val = num_prefix(val)
            data_sw['Total Rate NBM'] = "%.3g%s (cps)" % (val, unit_val)
            key_order_sw.append('Total Rate NBM')
        except (AttributeError, KeyError):
            pass
            
        # SE -----------------------------------------------------------------
        data_se = {'':''}
        key_order_se = []
            
        # get data: biases 
        try:
            if 'nqr_bias' in data.epics.keys():
                bias =      data.epics.nqr_bias.mean/1000.
                bias_std =  data.epics.nqr_bias.std/1000.
            elif 'nmr_bias' in data.epics.keys():
                bias =      data.epics.nmr_bias.mean
                bias_std =  data.epics.nmr_bias.std
            
            data_se["Platform Bias"] = "%.3f +/- %.3f kV" % \
                    (np.around(bias, 3), np.around(bias_std, 3))
            key_order_se.append("Platform Bias")
            
        except UnboundLocalError:
            pass
        
        try:
            data_se["BIAS15"] = "%.3f +/- %.3f V" % \
                    (np.around(data.epics.bias15.mean, 3), 
                     np.around(data.epics.bias15.std, 3))
            key_order_se.append('BIAS15')
        except AttributeError:
            pass
        
        # get data: beam energy
        try: 
            init_bias = data.epics.target_bias.mean
            init_bias_std = data.epics.target_bias.std
        except AttributeError:
            try:
                init_bias = data.epics.target_bias.mean
                init_bias_std = data.epics.target_bias.std
            except AttributeError:
                pass
            
        try:
            val = np.around(init_bias/1000., 3)
            std = np.around(init_bias_std/1000., 3)
            data_se["Initial Beam Energy"] = "%.3f +/- %.3f keV" % (val, std)
            key_order_se.append('Initial Beam Energy')
        except UnboundLocalError:
            pass
        
        # Get final beam energy
        try: 
            val = np.around(data.beam_kev, 3)
            std = np.around(data.beam_kev_err, 3)
            data_se['Implantation Energy'] = "%.3f +/- %.3f keV" % (val, std)
            key_order_se.append('Implantation Energy')
        except AttributeError:
            pass
        
        if key_order_se[-1] != '':
            key_order_se.append('')
        
        # doppler tubes
        try:
            data_se["Doppler Tube CH0"] = "%.3f +/- %.3f V" % \
                    (np.around(data.epics.dopplertube_ch0.mean, 3), 
                     np.around(data.epics.dopplertube_ch0.std, 3))
            key_order_se.append('Doppler Tube CH0')
        except AttributeError:
            pass
        
        try:
            data_se["Doppler Tube CH1"] = "%.3f +/- %.3f V" % \
                    (np.around(data.epics.dopplertube_ch1.mean, 3), 
                     np.around(data.epics.dopplertube_ch1.std, 3))
            key_order_se.append('Doppler Tube CH1')
        except AttributeError:
            pass
        
        try:
            data_se["Doppler Tube CH2"] = "%.3f +/- %.3f V" % \
                    (np.around(data.epics.dopplertube_ch2.mean, 3), 
                     np.around(data.epics.dopplertube_ch2.std, 3))
            key_order_se.append('Doppler Tube CH2')
        except AttributeError:
            pass
        
        if key_order_se[-1] != '':
            key_order_se.append('')
        
        # laser stuff
        try: 
            val = data.epics.las_pwr
            data_se['Laser Power'] = "%.3f +/- %.3f V" % (val.mean, val.std)
            key_order_se.append('Laser Power')
        except AttributeError:
            pass
        
        try: 
            val = data.epics.las_lambda
            data_se['Laser Wavelength'] = "%.3f +/- %.3f nm" % (val.mean, val.std)
            key_order_se.append('Laser Wavelength')
        except AttributeError:
            pass
        
        try: 
            val = data.epics.las_wavenum
            data_se['Laser Wavenumber'] = "%.5f +/- %.5f 1/cm" % (val.mean, val.std)
            key_order_se.append('Laser Wavenumber')
        except AttributeError:
            pass
        
        # magnet stuff
        try: 
            val = data.epics.hh_current.mean
            std = data.epics.hh_current.std
            field, _ = nqr_B0_hh3(amps=val)
            data_se['HH3 Magnet Current'] = "%.3f +/- %.3f A (%.2f G)" % (val, std, field)
            key_order_se.append('HH3 Magnet Current')            
        except AttributeError:
            pass
        
        try: 
            val = data.epics.hh6_current.mean
            std = data.epics.hh6_current.std
            field, _ = nqr_B0_hh6(amps=val)
            data_se['HH6 Magnet Current'] = "%.3f +/- %.3f A (%.2f G)" % \
                                                    (val, std, field)
            key_order_se.append('HH6 Magnet Current')            
        except AttributeError:
            pass
        
        if key_order_se[-1] != '':
            key_order_se.append('')
        
        # EL3
        try: 
            val = data.epics.el3.mean
            std = data.epics.el3.std
            data_se['Einzel Lens 3'] = "%.3f +/- %.3f V" % (val, std)
            key_order_se.append('Einzel Lens 3')
        except AttributeError:
            pass
        
        # NE -----------------------------------------------------------------
        data_ne = {'':''}
        key_order_ne = []
        
        # get data: SLR data
        if data.mode in ['20', '2h']:
            
            # PPG timing 
            try:
                dwell = int(data.ppg.dwelltime.mean)
                data_ne['Dwell time'] = "%d ms" % dwell
                key_order_ne.append('Dwell time')
            except AttributeError:
                pass
            
            try:    
                beam = int(data.ppg.prebeam.mean)            
                data_ne['Number of prebeam dwelltimes'] = "%d dwelltimes" % beam
                key_order_ne.append('Number of prebeam dwelltimes')
            except AttributeError:
                pass
            
            try:    
                beam = int(data.ppg.beam_on.mean)            
                data_ne['Number of beam-on dwelltimes'] = "%d dwelltimes" % beam
                key_order_ne.append('Number of beam-on dwelltimes')
            except AttributeError:
                pass
            
            try: 
                beam = int(data.ppg.beam_off.mean)
                data_ne['Number of beam-off dwelltimes'] = "%d dwelltimes" % beam
                key_order_ne.append('Number of beam-off dwelltimes')
            except AttributeError:
                pass
            
            try:    
                rf = int(data.ppg.rf_on_delay.mean)
                data_ne['RF on delay'] = "%d dwelltimes" % rf
                key_order_ne.append('RF on delay')
            except AttributeError:
                pass
            
            try:    
                rf = int(data.ppg.rf_on.mean)
                data_ne['RF on duration'] = "%d dwelltimes" % rf
                key_order_ne.append('RF on delay')
            except AttributeError:
                pass
            
            key_order_ne.append('')
            
            # Miscellaneous
            try:    
                hel = bool(data.ppg.hel_enable.mean)
                data_ne['Flip helicity'] = str(hel)
                key_order_ne.append('Flip helicity')
            except AttributeError:
                pass
            
            try:    
                hel = int(data.ppg.hel_sleep.mean)
                data_ne['Helicity sleep'] = "%d ms" % hel
                key_order_ne.append('Helicity sleep')
            except AttributeError:
                pass
        
            key_order_ne.append('')
            
            try:
                rf = bool(data.ppg.rf_enable.mean)
                data_ne['RF enable'] = str(rf)
                key_order_ne.append('RF enable')
                
                if rf:
                    freq = int(data.ppg.freq.mean)    
                    data_ne['Frequency'] = "%d Hz" % freq
                    key_order_ne.append('Frequency')
            except AttributeError:
                pass
            
        # get 1F specific data
        elif data.mode in ('1f', '1x'):
            
            # ppg timing box
            try:
                val = int(data.ppg.dwelltime.mean)
                data_ne['Bin Width'] = "%d ms" % val
                key_order_ne.append('Bin Width')
            except AttributeError:
                pass
            
            try:    
                val = int(data.ppg.nbins.mean)
                data_ne['Number of Bins'] = "%d" % val
                key_order_ne.append('Number of Bins')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.service_t.mean)
                data_ne['DAQ service time'] = "%d ms" % val
                key_order_ne.append('DAQ service time')
            except AttributeError:
                pass    
            
            key_order_ne.append('')
            
            # titles for frequencies
            if data.mode == '1x':
                nregions = len([k for k in data.ppg.keys() if 'fine_freq_start' in k])
                
                titles = ['Main scan']
                for i in range(nregions):
                    titles.append('Fine region %d' % (i+1))
                titles = [titles[i].ljust(14) for i in range(len(titles))]
                
                data_ne['Scan region'] = ''.join(titles)
                key_order_ne.append('Scan region')
            
            # PSM RF - Frequency
            try:
                start = int(data.ppg.freq_start.mean)
                data_ne['Start frequency'] = '%d Hz' % start
                key_order_ne.append('Start frequency')
            except AttributeError:
                pass
            
            try:
                stop = int(data.ppg.freq_stop.mean)
                data_ne['End frequency'] = '%d Hz' % stop
                key_order_ne.append('End frequency')
            except AttributeError:
                pass
            
            try:
                incr = int(data.ppg.freq_incr.mean)
                data_ne['Frequency scan gap'] = '%d Hz' % incr
                key_order_ne.append('Frequency scan gap')
            except AttributeError:
                pass
            
            try:
                nsteps = int((stop-start)/incr)
                data_ne['Num frequency steps'] = '%d (%d freqs in total)' % (nsteps, nsteps+1)
                key_order_ne.append('Num frequency steps')
            except NameError:
                pass
            
            try:
                center = (stop+start)/2
                data_ne['Center frequency'] = '%d Hz' % center
                key_order_ne.append('Center frequency')
            except NameError:
                pass
            
            try:
                width = stop-start
                data_ne['Full scan width'] = '%d Hz' % width
                key_order_ne.append('Full scan width')
            except NameError:
                pass
                
            try:
                val = bool(data.ppg.rand_freq_val.mean)
                data_ne['Randomize frequency order'] = str(val)
                key_order_ne.append('Randomize frequency order')
            except AttributeError:
                pass
                
            key_order_ne.append('')
            
            # Miscellaneous
            try:
                val = bool(data.ppg.hel_enable.mean)
                data_ne['Flip helicity'] = str(val)
                key_order_ne.append('Flip helicity')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.hel_sleep.mean)
                data_ne['Helicity sleep'] = "%d ms" % val
                key_order_ne.append('Helicity sleep')
            except AttributeError:
                pass
            
            try:
                val = bool(data.ppg.const_t_btwn_cycl.mean)
                data_ne['Ensure constant time between cycles'] = str(val)
                key_order_ne.append('Ensure constant time between cycles')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.ncycles.mean)
                data_ne['Repeat N cycles per supercycle'] = '%d' % val
                key_order_ne.append('Repeat N cycles per supercycle')
            except AttributeError:
                pass
                
            i = 1
            while True: 
                
                string_len = 14 * i
                
                try:
                    
                    start = int(data.ppg['fine_freq_start_%d' % i].mean)
                    string = data_ne['Start frequency'].ljust(string_len)
                    data_ne['Start frequency'] = string + '%d Hz' % start
                    
                    stop = int(data.ppg['fine_freq_end_%d' % i].mean)
                    string = data_ne['End frequency'].ljust(string_len)
                    data_ne['End frequency'] = string + '%d Hz' % stop
                    
                    gap = int(data.ppg['fine_freq_increment_%d' % i].mean)
                    string = data_ne['Frequency scan gap'].ljust(string_len) 
                    data_ne['Frequency scan gap'] = string + '%d Hz' % gap
                    
                    center = (start+stop)/2
                    string = data_ne['Center frequency'].ljust(string_len)
                    data_ne['Center frequency'] = string + '%d Hz' % center
                    
                    width = (stop-start)
                    string = data_ne['Full scan width'].ljust(string_len)
                    data_ne['Full scan width'] = string + '%d Hz' % width
                    
                    nsteps = int((stop-start)/gap)
                    string = data_ne['Num frequency steps'].split('(')[0]
                    string = string.ljust(string_len)
                    data_ne['Num frequency steps'] = string + '%d' % nsteps
                    
                except KeyError:
                    break
                except NameError:
                    pass
                
                i += 1
                
        # get 1W specific data
        elif data.mode == '1w':
            
            # ppg timing
            try:
                val = int(data.ppg.dwelltime.mean)
                data_ne['Bin width'] = "%d ms" % val
                key_order_ne.append('Bin width')
            except AttributeError:
                pass
            
            try:    
                val = int(data.ppg.nbins.mean)
                data_ne['Number of bins'] = "%d" % val
                key_order_ne.append('Number of bins')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.service_t.mean)
                data_ne['DAQ service time'] = "%d ms" % val
                key_order_ne.append('DAQ service time')
            except AttributeError:
                pass    
            
            
            key_order_ne.append('')
            
            # psm rf freq
            try:
                val = str(data.ppg.freqfn_f1.units)
                data_ne['Parametric function for f0ch1'] = val
                key_order_ne.append('Parametric function for f0ch1')
            except AttributeError:
                pass
            
            try:
                val = str(data.ppg.freqfn_f2.units)
                data_ne['Parametric function for f0ch2'] = val
                key_order_ne.append('Parametric function for f0ch2')
            except AttributeError:
                pass
            
            try:
                val = str(data.ppg.freqfn_f3.units)
                data_ne['Parametric function for f0ch3'] = val
                key_order_ne.append('Parametric function for f0ch3')
            except AttributeError:
                pass
            
            try:
                val = str(data.ppg.freqfn_f4.units)
                data_ne['Parametric function for f0ch4'] = val
                key_order_ne.append('Parametric function for f0ch4')
            except AttributeError:
                pass
             
            try:
                val = int(data.ppg.yconst.mean)
                data_ne['Parametric Y constant'] = '%d Hz' % val
                key_order_ne.append('Parametric Y constant')
            except AttributeError:
                pass
                
            try:
                start = int(data.ppg.xstart.mean)
                data_ne['Parametric X start'] = '%d Hz' % start
                key_order_ne.append('Parametric X start')
            except AttributeError:
                pass
                
            try:
                stop = int(data.ppg.xstop.mean)
                data_ne['Parametric X end'] = '%d Hz' % stop
                key_order_ne.append('Parametric X end')
            except AttributeError:
                pass
                
            try:
                incr = int(data.ppg.xincr.mean)
                data_ne['X scan gap'] = '%d Hz' % incr
                key_order_ne.append('X scan gap')
            except AttributeError:
                pass
                
            try:
                nsteps = int((stop-start)/incr)
                data_ne['Num X steps'] = '%d (%d freqs total)' % (nsteps, nsteps+1)
                key_order_ne.append('Num X steps')
            except NameError:
                pass

            try:
                val = bool(data.ppg.rand_freq_val.mean)
                data_ne['Randomize frequency order'] = str(val)
                key_order_ne.append('Randomize frequency order')
            except AttributeError:
                pass
                
            key_order_ne.append('')
                
            # Miscellaneous
            try:
                val = bool(data.ppg.hel_enable.mean)
                data_ne['Flip helicity'] = str(val)
                key_order_ne.append('Flip helicity')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.hel_sleep.mean)
                data_ne['Helicity sleep'] = "%d ms" % val
                key_order_ne.append('Helicity sleep')
            except AttributeError:
                pass        
            
            try:
                val = bool(data.ppg.const_t_btwn_cycl.mean)
                data_ne['Ensure constant time between cycles'] = str(val)
                key_order_ne.append('Ensure constant time between cycles')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.ncycles.mean)
                data_ne['Repeat N cycles per supercycle'] = '%d' % val
                key_order_ne.append('Repeat N cycles per supercycle')
            except AttributeError:
                pass
            
        # get Rb Cell specific data
        elif data.mode in ['1n', '1d', '1e', '1c']:
            
            try:
                dwell = int(data.ppg.dwelltime.mean)
                data_ne['Bin Width'] = "%d ms" % dwell
                key_order_ne.append('Bin Width')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.nbins.mean)
                data_ne['Number of Bins'] = '%d' % val
                key_order_ne.append('Number of Bins')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.service_t.mean)
                data_ne['DAQ service time'] = "%d ms" % val
                key_order_ne.append('DAQ service time')
            except AttributeError:
                pass    
            
            data_ne[''] = ''
            key_order_ne.append('')
            
             # common items
            try:
                val = bool(data.ppg.hel_enable.mean)
                data_ne['Flip helicity'] = str(val)
                key_order_ne.append('Flip helicity')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.hel_sleep.mean)
                data_ne['Helicity sleep'] = "%d ms" % val
                key_order_ne.append('Helicity sleep')
            except AttributeError:
                pass
            
            # custom varible scan check
            try: 
                custom_enable = bool(data.ppg.customv_enable.mean)
            except AttributeError:
                custom_enable = False
            
            # set scan variable name, ppg key
            if data.mode == '1c':
                prefix = ''
                try:
                    val = str(data.ppg.scan_device.units)
                    data_ne['CAMP device'] = '%s' % val
                    key_order_ne.append('CAMP device')
                except AttributeError:
                    pass                
                
            else:
                if custom_enable:
                    prefix = 'customv_'
                    
                    try:
                        val = str(data.ppg.customv_name_write.units)
                        data_ne['EPICS device'] = '%s' % val
                        key_order_ne.append('EPICS device')
                    except AttributeError:
                        pass                
                else:
                    prefix = ''
                    data_ne['EPICS device'] = '%s' % self.mode_epics_var[self.data.mode]
                    key_order_ne.append('EPICS device')                
                    
            # scan ranges
            try:
                ppg = data.ppg[prefix+'scan_start']
                start = int(ppg.mean)
                unit = str(ppg.units).title()
                data_ne['Scan start'] = '%d %s' % (start, unit)
                key_order_ne.append('Scan start')
            except (AttributeError, KeyError):
                pass
            
            try:    
                ppg = data.ppg[prefix+'scan_stop']
                stop = int(ppg.mean)
                unit = str(ppg.units).title()
                data_ne['Scan stop'] = '%d %s' % (stop, unit)
                key_order_ne.append('Scan stop')
            except (AttributeError, KeyError):
                pass
            
            try:
                ppg = data.ppg[prefix+'scan_incr']
                incr = int(np.round(ppg.mean))
                unit = str(ppg.units).title()
                data_ne['Scan gap'] = '%d %s' % (incr, unit)
                key_order_ne.append('Scan gap')
            except (AttributeError, KeyError):
                pass
            
            try:
                nsteps = int((stop-start)/incr)
                data_ne['Number of scan steps'] = '%d (%d data points)' % (nsteps, nsteps+1)
                key_order_ne.append('Number of scan steps')
            except NameError:
                pass
                
            try:
                val = int(data.ppg.ncycles.mean)
                data_ne['Repeat N cycles per supercycle'] = '%d' % val
                key_order_ne.append('Repeat N cycles per supercycle')
            except AttributeError:
                pass
            
        # get 2e mode specific data
        elif data.mode in ['2e']:
            
            # ppg timing
            try:
                val = int(data.ppg.rf_on_ms.mean)
                data_ne['RF on time (aka dwelltime)'] = "%d ms" % val
                key_order_ne.append('RF on time (aka dwelltime)')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.rf_on_delay.mean)
                data_ne['RF on delay'] = "%d dwelltimes" % val
                key_order_ne.append('RF on delay')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.beam_off_ms.mean)
                data_ne['Beam off time'] = "%d ms" % val
                key_order_ne.append('Beam off time')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.ndwell_post_on.mean)
                data_ne['Post-RF beam-on'] = "%d dwelltimes" % val
                key_order_ne.append('Post-RF beam-on')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.ndwell_per_f.mean)
                data_ne['Number dwelltimes per freq'] = "%d" % val
                key_order_ne.append('Number dwelltimes per freq')
            except AttributeError:
                pass
                
            key_order_ne.append('')
            
            # PSM RF Frequency   
            try:
                start = int(data.ppg.freq_start.mean)
                data_ne['Start frequency'] = '%d Hz' % start
                key_order_ne.append('Start frequency')
            except AttributeError:
                pass
            
            try:
                stop = int(data.ppg.freq_stop.mean)
                data_ne['End frequency'] = '%d Hz' % stop
                key_order_ne.append('End frequency')
            except AttributeError:
                pass
            
            try:
                incr = int(data.ppg.freq_incr.mean)
                data_ne['Frequency scan gap'] = '%d Hz' % incr
                key_order_ne.append('Frequency scan gap')
            except AttributeError:
                pass
            
            try:
                nsteps = int((stop-start)/incr)
                data_ne['Num frequency steps'] = '%d (%d freqs in total)' % (nsteps, nsteps+1)
                key_order_ne.append('Num frequency steps')
            except NameError:
                pass
            
            try:
                center = (stop-start)/2
                data_ne['Center frequency'] = '%d Hz' % center
                key_order_ne.append('Center frequency')
            except NameError:
                pass
            
            try:
                width = stop-start
                data_ne['Full scan width'] = '%d Hz' % width
                key_order_ne.append('Full scan width')
            except NameError:
                pass
                
            try:
                val = bool(data.ppg.rand_freq_val.mean)
                data_ne['Randomize frequency order'] = str(val)
                key_order_ne.append('Randomize frequency order')
            except AttributeError:
                pass
                
            key_order_ne.append('')
            
            # misc
            try:
                val = bool(data.ppg.hel_enable.mean)
                data_ne['Flip helicity'] = str(val)
                key_order_ne.append('Flip helicity')
            except AttributeError:
                pass
                
            try:
                val = bool(data.ppg.hel_enable.mean)
                data_ne['Helicity sleep'] = "%d ms" % val
                key_order_ne.append('Helicity sleep')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.ncycles.mean)
                data_ne['Repeat N cycles per supercycle'] = '%d' % val
                key_order_ne.append('Repeat N cycles per supercycle')
            except AttributeError:
                pass
            
        # set viewer string
        def set_str(data_dict, key_order, txtbox):
        
            m = max(max(map(len, list(data_dict.keys()))) + 1, 5)
            lines = [k.rjust(m)+':   ' + data_dict[k] for k in key_order]
            lines = [l if l.strip() != ':' else '' for l in lines]
            
            self.set_textbox_text(txtbox, '\n'.join(lines))
        
        set_str(data_nw, key_order_nw, self.text_nw)
        set_str(data_ne, key_order_ne, self.text_ne)
        set_str(data_sw, key_order_sw, self.text_sw)
        set_str(data_se, key_order_se, self.text_se)
        
        return True
        
    # ======================================================================= #
    def _get_latest_run(self, year, run):
        """
            Get run number of latest run in local file system, given an initial 
            part of the run number
        """
        
        runlist = []
            
        # look for latest run by run number
        for d in [self.bfit.bnmr_archive_label, self.bfit.bnqr_archive_label]:
            if d not in os.environ.keys():
                if 'BNMR' in d.upper():
                    dirloc = os.path.join(bd._mud_data, 'bnmr')
                else:
                    dirloc = os.path.join(bd._mud_data, 'bnqr')
            else:
                dirloc = os.environ[d]
            runlist.extend(glob.glob(os.path.join(dirloc, str(year), '0%d*.msr'%run)))
        runlist = [int(os.path.splitext(os.path.basename(r))[0]) for r in runlist]
        
        # get latest run by max run number
        try:
            run = max(runlist)
        except ValueError:
            self.logger.exception('Run fetch failed')
            for t in [self.text_nw, self.text_ne, self.text_sw, self.text_se]:
                self.set_textbox_text(t, 'Run not found.')  
            return False
        else:
            return run
        
    # ======================================================================= #
    def draw(self, figstyle, quiet=False):
        """Get data then draw."""
        self.bfit.logger.info('Draw button pressed')
        
        if self.get_data(quiet=quiet):
            self.data.draw(self.asym_type.get(), 
                           label=self.bfit.get_label(self.data), 
                           figstyle=figstyle,
                           asym_args={  
                                'slr_bkgd_corr':self.bfit.correct_bkgd.get(),
                                'slr_rm_prebeam':not self.bfit.draw_prebin.get()})
            
    # ======================================================================= #
    def draw_diagnostics(self): #incomplete
        """
            Get data then draw in debug mode.
        """
        
        # isssue with data fetch
        if not self.get_data(quiet=quiet):
            return
        
        # get data
        dat = self.data
    
        # make figure
        fix, (ax1, ax2, ax3, ax4) = plt.subplots(nrows=2, ncols=2)
    
        # get asym
        a = data.asym(hist_select=self.bit.hist_select)
        x = a[self.x_tag[data.mode]]
        xlabel = self.xlabel_dict[data.mode]
            
        # draw 2e mode 
        if '2e' == dat.mode:
            pass 
            
        # draw TD mode
        elif '2' in dat.mode:
            
            # draw combined asym -------------------------
            tag = a.c[0]!=0 # remove zero asym
            ax1.errorbar(x[tag], a.c[0][tag], a.c[1][tag])
            ax1.set_xlabel(xlabel)
            ax1.set_ylabel(self.bfit.ylabel_dict['c'])
            
            # draw split asym ----------------------------
            
            # remove zero asym
            ap = a.p[0]
            an = a.n[0]
            tag_p = ap!=0
            tag_n = an!=0
            tag_cmb = tag_p*tag_n
            
            # get average
            avg = np.mean(ap[tag_cmb]+an[tag_cmb])/2
            
            # draw
            ax2.errorbar(x[tag_p], ap[tag_p], a.p[1][tag_p], label='+')
            ax2.errorbar(x[tag_n], an[tag_n], a.n[1][tag_n], label="-")
            ax2.axhline(avg, color='k', linestyle='--')
            ax2.set_xlabel(xlabel)
            ax2.set_ylabel(self.bfit.ylabel_dict['h'])
        
            # draw histograms  --------------------------
            hist = data.hist
            
            # draw
            keylist = ('F+', 'F-', 'B+', 'B-', 'L+', 'R+', 'L-', 'R-', 
                         'NBMF+', 'NBMF-', 'NBMB+', 'NBMB-', 'AL0+', 'AL0-')
            for i, h in enumerate(keylist):
                
                # get bins
                try:
                    x = np.arange(len(hist[h].data))
                except KeyError:
                    continue
                
                # check for non-empty histograms, then draw
                if np.mean(hist[h].data) > 0:                        
                    ax3.plot(x, hist[h].data, label=h)
                    
            ax3.ylabel(self.bfit.ylabel_dict['rhist'])
            ax3.xlabel('Bin')
        
        # draw TI mode
        elif '1' in dat.mode:
            pass
        
        # unknown mode
        else:
            raise RuntimeError('Unknown mode type')
    
    # ======================================================================= #
    def export(self, filename=None):
        """Export data as csv"""
        
        self.logger.info('Export button pressed')
        
        # get data
        if not self.get_data():
            return
        data = self.data
        
        # get filename 
        if filename is None:
            filename = filedialog.asksaveasfilename(
                initialfile=self.default_export_filename%(data.year, data.run), 
                filetypes=[('csv', '*.csv'), 
                           ('allfiles', '*')], 
                defaultextension='.csv')
        
        # write to file
        if filename:
            self.bfit.export(data, filename, rebin=self.rebin.get())
    
    # ======================================================================= #
    def get_data(self, quiet=False):
        """
            Display data and send bdata object to bfit draw list. 
            Return True on success, false on Failure
        """
        
        if self.filename != '':
            return self._get_data(bdata(0, filename=self.filename))
        
        # fetch year
        try:
            year = self.year.get()
        except ValueError:
            for t in [self.text_nw, self.text_ne, self.text_sw, self.text_se]:
                self.set_textbox_text(t, 'Year input must be integer valued')  
                self.logger.exception('Year input must be integer valued')
            return False
        
        # fetch run number
        run = self.runn.get()
        
        self.logger.debug('Parsing run input %s', run)
        
        if run < 40000:
            run = self._get_latest_run(year, run)
            if run is False:
                return False
        
        # check for merged runs
        if run > 50000:
            
            # split run number
            run = str(run)
            run = [int(run[i:i+5]) for i in range(0, len(run), 5)]
            
            # split year number - note to self: update in 1000 years
            if year > 3000:
                year = str(year)
                year = [int(year[i:i+4]) for i in range(0, len(year), 4)]
            else:
                year = [year]*len(run)
            
            # get data
            data = bmerged([bdata(r, year=y) for r, y in zip(run, year)])
            
        else:
            data = bdata(run, year=year)
        
        self.logger.info('Fetching run %s from %s', run, year)
        
        return self._get_data(data, quiet=quiet)
        
    # ======================================================================= #
    def load_file(self, filename=None):
        """
            Read data based on filename rather than run number
            
            return True on success and False on failure
        """
        
        # get filename
        if filename is None:
            
            if self.filename == '':
                filename = filedialog.askopenfilename(filetypes=[('msr', '*.msr'),
                                                             ('allfiles', '*')])
                if not filename:
                    return False    
            else:
                filename = ''
        
        # set filename
        self.logger.debug('self.filename = %s', filename)
        self.filename = filename
        
        # get data
        if filename != '':
            self.logger.info('Fetching run %s', filename)
            self.button_run_from_file['text'] = 'Stop loading run from file'
            self.entry_year['state'] = 'disabled'
            self.entry_runn['state'] = 'disabled'
            return self._get_data(bdata(0, filename=filename))
        else:
            self.button_run_from_file['text'] = 'Load run from file'
            self.entry_year['state'] = 'normal'
            self.entry_runn['state'] = 'normal'
            return True
                    
    # ======================================================================= #
    def set_rebin(self, data):
        """
            Link rebin intvar to that of data object
            data: fitdata obj
        """
        value = self.rebin.get()
        self.rebin = data.rebin
        self.rebin.set(value)
        self.entry_rebin.configure(textvariable = data.rebin)
        
    # ======================================================================= #
    def set_nbm(self):
        """
            Set the nbm variable based on the run mode
        """
        
        # check if data
        if not hasattr(self, 'data'):
            return
        
        # check run mode
        mode = self.data.mode
        
        self.bfit.set_nbm(mode)

    # ======================================================================= #
    def set_textbox_text(self, textbox, text):
        """Set the text in a tkinter Text widget"""
        textbox.delete('1.0', END)
        textbox.insert('1.0', text)
        
    # ======================================================================= #
    def do_update(self, first=True, runid=''):
        self.logger.debug('Draw via periodic update')
        
        # update stop condition
        if runid and runid != self.update_id:
            return
        
        # select period drawing figure
        if first:
            
            first = False
            
            # check that there is a canvas, if not, draw
            if self.bfit.plt.active['inspect'] == 0:
                self.draw('inspect', quiet=False)
                first = True
            
            # set up updating canvas
            fig = self.bfit.plt.gcf('inspect')
            fig.canvas.manager.set_window_title('Figure %d (Inspect - Updating)'%fig.number)
            self.bfit.plt.plots['periodic'] = [fig.number]
            self.bfit.plt.active['periodic'] = self.bfit.plt.active['inspect']
            
            runid = self.data.id
            self.update_id = runid
            
            # repeat
            if not first:
                self.bfit.root.after(self.bfit.update_period*1000, 
                                     lambda:self.do_update(first=False, 
                                                           runid=runid))
                return
        
        # update 
        if self.is_updating.get():
            
            # check that figure exists or is not updating (was closed)
            if self.bfit.plt.active['periodic'] not in self.bfit.plt.plots['inspect']: 
                self.is_updating.set(False)
                del self.bfit.plt.plots['periodic'][0]
                self.bfit.plt.active['periodic'] = 0
                return
            
            # Get the updating figure
            fig = self.bfit.plt.gcf('periodic')
            title = fig.canvas.manager.get_window_title()

            # Check that the figure is still updating
            if 'Updating' not in title:
                self.is_updating.set(False)
                del self.bfit.plt.plots['periodic'][0]
                self.bfit.plt.active['periodic'] = 0
                return
            
            # update run
            year, run = tuple(map(int, runid.split('.')[:2]))
            current_year = self.year.get()
            current_run = self.runn.get()
            
            self.year.set(year)
            self.runn.set(run)
            
            # check current run 
            if current_run < 40000:
                current_run2 = self._get_latest_run(current_year, current_run)
                if current_run2 is False:
                    return 
            else:
                current_run2 = current_run
                
            # update only in stack mode
            draw_style = self.bfit.draw_style.get()
            self.bfit.draw_style.set('stack')
            
            # don't update figure limits
            if self.data.mode in ('20', '2h'):
                self.bfit.plt.autoscale('periodic', False)
                
            # draw
            self.draw(figstyle='periodic', quiet=True)
            draw_style = self.bfit.draw_style.set(draw_style)
            
            # reset year and run 
            do_quiet = (current_run2 != run) or (current_year != year)
            
            self.year.set(current_year)
            self.runn.set(current_run)
            self.get_data(quiet=do_quiet)
            
            # Print update message
            print('Updated figure at:', str(datetime.datetime.now()).split('.')[0], 
                  flush=True)
            
            # repeat
            self.bfit.root.after(self.bfit.update_period*1000, 
                                 lambda:self.do_update(first=False, runid=runid))
            
        # remove window from updating list
        else:
            # check if window already removed 
            if self.bfit.plt.active['periodic'] != 0:
                
                # remove window
                fig = self.bfit.plt.gcf('periodic')
                fig.canvas.manager.set_window_title('Figure %d (Inspect)'%fig.number)
                
                self.bfit.plt.autoscale('periodic', True)
                del self.bfit.plt.plots['periodic'][0]
                self.bfit.plt.active['periodic'] = 0
            
# =========================================================================== #
def num_prefix(val):
    if val > 1e9: return (val/1e9, 'G')
    if val > 1e6: return (val/1e6, 'M')
    if val > 1e3: return (val/1e3, 'k')
    
    return (val, '')



# Test constrained fit module
from numpy.testing import *
import numpy as np
from bfit.gui.popup_fit_constraints import popup_fit_constraints
from bfit.gui.bfit import bfit

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')

def with_bfit(function):
    
    def wrapper(*args, **kwargs):
        # make gui
        b = bfit(None, True)
        tab = b.fit_files
        tab2 = b.fetch_files
        b.notebook.select(2)
        
        tab2.year.set(2020)
        tab2.run.set('40123 40124')
        tab2.get_data()
        tab.populate()
        b.draw_fit.set(False)
        
        try:
            return function(*args, **kwargs, b=b, fittab=tab, fetchtab=tab2)
        finally:
            b.on_closing()
            del b
            
    return wrapper

def check_line_names(fittab, names):
    names = sorted(names)
    for fline in fittab.fit_lines.values():
        pname = sorted([line.pname for line in fline.lines])
        assert pname == names, 'incorrect pnames in run {id}'.format(id=fline.data.id)

def check_line_state(fittab, disabled_pnames):
    for fline in fittab.fit_lines.values():
        states = {line.pname: line.entry['p0']['state'] for line in fline.lines}
        
        for pname, state in states.items():
            
            if state == 'normal':
                assert pname not in disabled_pnames, \
                        '{pname} is disabled when it should not be'.format(pname=pname)
            elif state == 'disabled':
                assert pname in disabled_pnames, \
                        '{pname} is not disabled when it should be'.format(pname=pname)
            else:
                raise RuntimeError('Unknown run state %s' % state)

def check_line_p0(fittab):
    # make sure all lines have p0 filled in (bounds should also be filled in if p0 is)
    
    for fline in fittab.fit_lines.values():
        for line in fline.lines:
            assert line.entry['p0'].get() != '', \
                '"{pname}" of {run} has no p0'.format(pname=line.pname, run=fline.data.id)

@with_bfit
def test_par_detection(b=None, fittab=None, fetchtab=None):
    
    # start up
    fittab.show_constr_window()
    constr = fittab.pop_fitconstr
    
    # check input on first attempt
    constr.entry.insert('1.0', 'amp = a*np.exp(b*BIAS**0.5)')
    constr.get_input()
    
    assert constr.defined == ['amp'], 'defined parameters incorrect'
    assert constr.eqn == ['a*np.exp(b*BIAS**0.5)'], 'eqn incorrect'
    assert constr.new_par == [['a', 'b']], 'new_par incorrect'
    assert constr.new_par_unique == ['a', 'b'], 'new_par_unique incorrect'
    
    # change input
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = d+f')
    constr.get_input()
    
    assert constr.defined == ['amp'], 'defined parameters incorrect'
    assert constr.eqn == ['d+f'], 'eqn incorrect'
    assert constr.new_par == [['d', 'f']], 'new_par incorrect'
    assert constr.new_par_unique == ['d', 'f'], 'new_par_unique incorrect'
    
    # add defined input
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = d+f\n1_T1 = a')
    constr.get_input()
    
    assert constr.defined == ['amp', '1_T1'], 'defined parameters incorrect'
    assert constr.eqn == ['d+f', 'a'], 'eqn incorrect'
    assert constr.new_par == [['d', 'f'], ['a']], 'new_par incorrect'
    assert constr.new_par_unique == ['a', 'd', 'f'], 'new_par_unique incorrect'
    
    # test no new parameters
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = 1')
    constr.get_input()
    
    assert constr.defined == ['amp'], 'defined parameters incorrect'
    assert constr.eqn == ['1'], 'eqn incorrect'
    assert constr.new_par == [[]], 'new_par incorrect'
    assert constr.new_par_unique == [], 'new_par_unique incorrect'

@with_bfit
def test_set_constr_lines(b=None, fittab=None, fetchtab=None):

    # start up
    fittab.show_constr_window()
    constr = fittab.pop_fitconstr
    
    # check good input
    constr.entry.insert('1.0', 'amp = a+b')
    constr.get_input()
    constr.set_constraints()
    check_line_names(fittab, ['1_T1', 'amp', 'a', 'b'])
    
    # check change n new_par
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = a')
    constr.get_input()
    constr.set_constraints()
    check_line_names(fittab, ['1_T1', 'amp', 'a'])
    
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = a+c')
    constr.get_input()
    constr.set_constraints()
    check_line_names(fittab, ['1_T1', 'amp', 'a', 'c'])
    
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', '')
    constr.get_input()
    constr.set_constraints()
    check_line_names(fittab, ['1_T1', 'amp'])

@with_bfit
def test_disable(b=None, fittab=None, fetchtab=None):
    """
        check init button and line disable
    """
    
    # start up
    constr = fittab.pop_fitconstr
    
    # a and b
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = a+b')
    constr.get_input()
    constr.set_constraints()
    check_line_state(fittab, ['amp'])
    
    # remove a
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = b')
    constr.get_input()
    constr.set_constraints()
    check_line_state(fittab, ['amp'])
    
    # switch to T1
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', '1_T1 = b')
    constr.get_input()
    constr.set_constraints()
    check_line_state(fittab, ['1_T1'])

    # none
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', '')
    constr.get_input()
    constr.set_constraints()
    check_line_state(fittab, [])
    
    # check force modify
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = a')
    constr.get_input()
    constr.set_constraints()
    fittab.populate_param(force_modify=True)
    check_line_state(fittab, [])
    
@with_bfit
def test_adding_runs(b=None, fittab=None, fetchtab=None):
    """
        Test changing which runs are selected or not
    """
    # start up
    constr = fittab.pop_fitconstr
    
    # a and b as normal
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = a+b')
    constr.get_input()
    constr.set_constraints()
    check_line_state(fittab, ['amp'])
    check_line_names(fittab, ['1_T1', 'amp', 'a', 'b'])
    check_line_p0(fittab)
    
    # uncheck one run
    b.data['2020.40124'].check_state.set(False)
    fittab.populate()
    check_line_state(fittab, ['amp'])
    check_line_names(fittab, ['1_T1', 'amp', 'a', 'b'])
    check_line_p0(fittab)
    
    # repopulate
    fittab.populate_param(force_modify=True)
    check_line_state(fittab, [])
    check_line_names(fittab, ['1_T1', 'amp'])
    check_line_p0(fittab)
    
    # add back the run
    b.data['2020.40124'].check_state.set(True)
    fittab.populate()
    check_line_state(fittab, [])
    check_line_names(fittab, ['1_T1', 'amp'])
    check_line_p0(fittab)
    
    # uncheck run, add constraints
    b.data['2020.40124'].check_state.set(False)
    fittab.populate()
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = a+b')
    constr.get_input()
    constr.set_constraints()
    check_line_state(fittab, ['amp'])
    check_line_names(fittab, ['1_T1', 'amp', 'a', 'b'])
    check_line_p0(fittab)
    
    # add back the run
    b.data['2020.40124'].check_state.set(True)
    fittab.populate()
    check_line_state(fittab, ['amp'])
    check_line_names(fittab, ['1_T1', 'amp', 'a', 'b'])
    check_line_p0(fittab)
    
    # change the data type to 1f
    fetchtab.remove_all()
    fetchtab.year.set(2021)
    fetchtab.run.set('40123')
    fetchtab.get_data()
    fittab.populate()
    
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = a')
    constr.get_input()
    constr.set_constraints()
    
    fetchtab.remove_all()
    fetchtab.run.set('40033')
    fetchtab.get_data()
    fittab.populate()
    
    check_line_state(fittab, [])
    check_line_names(fittab, ['baseline', 'fwhm', 'height', 'peak'])
    check_line_p0(fittab)
        
@with_bfit
def test_fit(b=None, fittab=None, fetchtab=None):
    
    # start up
    constr = fittab.pop_fitconstr
    
    # fit without constraints
    fittab.do_fit()
    result = b.data['2020.40123'].fitpar.copy()
    
    # fit with simple constrained identity
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = a')
    constr.get_input()
    constr.set_constraints()    
    fittab.do_fit()
    
    new_result = b.data['2020.40123'].fitpar
    
    assert_almost_equal(result.loc['amp', 'res'], new_result.loc['amp', 'res'], 
                        err_msg = 'identity fit result not equal', decimal=5)
    
    assert_almost_equal(result.loc['amp', 'res'], new_result.loc['a', 'res'], 
                        err_msg = 'identity fit result not equal', decimal=5)
    
    # fit with simple function constraint
    fittab.show_constr_window()
    constr.entry.delete('1.0', 'end')
    constr.entry.insert('1.0', 'amp = a*2')
    constr.get_input()
    constr.set_constraints()    
    fittab.do_fit()
    
    new_result = b.data['2020.40123'].fitpar
    
    assert_almost_equal(result.loc['amp', 'res'], new_result.loc['amp', 'res'], 
                        err_msg = 'identity fit result not equal', decimal=2)
    
    assert_almost_equal(result.loc['amp', 'res'], new_result.loc['a', 'res']*2, 
                        err_msg = 'identity fit result not equal', decimal=2)
    
    



# test inspect tab
# Derek Fujimoto
# Feb 2021

from numpy.testing import *
import numpy as np
from bfit.gui.calculator_nqr_B0 import calculator_nqr_B0

def test_calculate_voltage():
    calc = calculator_nqr_B0(True)
    
    # voltage
    calc.current.set('1')
    calc.entry_current.focus_set()
    calc.calculate()
    assert_almost_equal(float(calc.field.get()), 2.3878, 
        err_msg = "calculator nqr B0 set current = 1, compared to online calculator")
    calc.root.destroy()
    
def test_calculate_field():
    calc = calculator_nqr_B0(True)
    
    # field
    calc.field.set('1')
    calc.entry_field.focus_set()
    calc.calculate()
    assert_almost_equal(float(calc.current.get()), 0.3729, 
                        err_msg = "calculator nqr B0 set field = 1, compared to online calculator")
    calc.root.destroy()




# Model the fit results with a function
# Derek Fujimoto
# August 2019

from tkinter import *
from tkinter import ttk, messagebox

import logging, re, os, warnings
import weakref as wref
import numpy as np
import pandas as pd

from bfit import logger_name
from bfit.backend.entry_color_set import on_focusout, on_entry_click
from bfit.backend.raise_window import raise_window
import bfit.backend.colors as colors

# ========================================================================== #
class template_fit_popup(object):
    """
        Base class for fitting popup windows

        bfit
        fittab
        logger

        entry:              Text, text entry for user
        entry_label:        Label, instructions above input text box
        input_fn_text:      string, text defining input functions
        left_frame:         Frame, for additional details


        output_par_text     text, detected parameter names
        output_text         dict, keys: p0, blo, bhi, res, err, value: tkk.Text objects

        output_par_text_val string, contents of output_par_text

        reserved_pars:      dict, define values in bdata that can be accessed
        right_frame:        Frame, for input and actions
        win:                Toplevel
    """

    # default parameter values on new parameter
    default_parvals = { 'p0':1,
                        'blo':-np.inf,
                        'bhi':np.inf,
                        'res':np.nan,
                        'err-':np.nan,
                        'err+':np.nan}

    window_title = 'Base class popup window'

    # ====================================================================== #
    def __init__(self, bfit, input_fn_text=''):

        self.bfit = bfit
        self.input_fn_text = input_fn_text

        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')

        # draw window
        self.logger.debug('Initialization success. Starting mainloop.')

    # ====================================================================== #
    def show(self):
        """
            Show window
        """
        bfit = self.bfit

        # make a new window
        self.win = Toplevel(bfit.mainframe)
        self.win.title(self.window_title)
        frame = ttk.Frame(self.win, relief='sunken', pad=5)
        left_frame = ttk.Frame(frame)
        right_frame = ttk.Frame(frame)

        # set icon
        self.bfit.set_icon(self.win)

        # Key bindings
        self.win.bind('<Control-Key-Return>', self.do_return)
        self.win.bind('<Control-Key-KP_Enter>', self.do_return)

        # Text entry
        entry_frame = ttk.Frame(right_frame, relief='sunken', pad=5)
        self.entry_label = ttk.Label(entry_frame, justify=LEFT,
                                text='')
        self.entry = Text(entry_frame, width=60, height=13, state='normal')
        self.entry.bind('<KeyRelease>', self.get_input)
        scrollb = Scrollbar(entry_frame, command=self.entry.yview)
        self.entry['yscrollcommand'] = scrollb.set

        # Insert default text
        self.entry.insert('1.0', self.input_fn_text.strip())

        # gridding
        scrollb.grid(row=1, column=1, sticky='nsew')
        self.entry_label.grid(column=0, row=0, sticky=W)
        self.entry.grid(column=0, row=1)

        # grid to frame
        frame.grid(column=0, row=0)
        left_frame.grid(column=0, row=0, sticky=(N, S))
        right_frame.grid(column=1, row=0, sticky=(N, S))
        entry_frame.grid(column=0, row=0, sticky=(N, E, W), padx=1, pady=1)

        # initialize
        self.left_frame = left_frame
        self.right_frame = right_frame

    # ====================================================================== #
    def cancel(self):
        self.win.destroy()

    # ====================================================================== #
    def do_return(self, *_):
        """
            Activated on press of return key
        """
        pass

    # ====================================================================== #
    def do_parse(self, *_):
        """
            Detect new global variables
            returns split lines, new parameter names
        """

        # clean input
        text = self.input_fn_text.split('\n')
        text = [t.strip() for t in text if '=' in t]

        # check for no input
        if not text:
            return (None, None, None)

        # get equations and defined variables
        defined = [t.split('=')[0].strip() for t in text]
        eqn = [t.split('=')[1].strip() for t in text]

        # check for new parameters
        new_par = []
        for eq in eqn:
            lst = re.split('\W+', eq)    # split list non characters

            # throw out known things: numbers numpy equations
            delist = []
            for i, l in enumerate(lst):

                # check numpy functions
                if l == 'np':
                    delist.append(i)
                    delist.append(i+1)
                    continue

                # check integer
                try:
                    int(l)
                except ValueError:
                    pass
                else:
                    delist.append(i)
                    continue

                # check variables
                if l in self.reserved_pars:
                    delist.append(i)
                    continue

            delist.sort()
            for i in delist[::-1]:
                try:
                    del lst[i]
                except IndexError:  # error raised on incomplete math: ex "np."
                    pass

            # look for non-parameters
            lst = [i for i in lst if i]

            new_par.append(lst)

        # logging
        self.logger.info('Parse found constraints for %s, and defined %s',
                         sorted(defined),
                         new_par)

        return (defined, eqn, new_par)

    # ====================================================================== #
    def do_after_parse(self, defined=None, eqn=None, new_par=None):
        pass

    # ====================================================================== #
    def get_input(self, *_):
        """Get input from text box."""
        self.input_fn_text = self.entry.get('1.0', END)
        out = self.do_parse()
        self.do_after_parse(*out)

    # ====================================================================== #
    def yview(self, *args):
        """
            Scrollbar for all output text fields
        """
        self.output_par_text.yview(*args)
        for k in self.output_text:
            self.output_text[k].yview(*args)



# test inspect tab
# Derek Fujimoto
# Feb 2021

from numpy.testing import *
import numpy as np
import matplotlib.pyplot as plt
from bfit.gui.bfit import bfit

# filter unneeded warnings
import pytest
pytestmark = pytest.mark.filterwarnings('ignore:2020')

def with_bfit(function):
    
    def wrapper(*args, **kwargs):
        # make gui
        b = bfit(None, True)
        tab = b.fetch_files
        b.notebook.select(1)
        
        try:
            function(*args, **kwargs, tab=tab, b=b)
        finally:
            b.on_closing()
            del b
            
    return wrapper

@with_bfit
def test_fetch(tab=None, b=None):
    
    # set year
    tab.year.set(2020)
    
    # get one
    tab.run.set('40123')
    tab.get_data()
    assert_equal(len(list(tab.data_lines.keys())), 1, 'fetch tab fetch single run')
    
    tab.run.set('40124')
    tab.get_data()
    assert_equal(len(list(tab.data_lines.keys())), 2, 'fetch tab fetch another single run')
    
    # get two
    tab.run.set('40125 40126')
    tab.get_data()
    assert_equal(len(list(tab.data_lines.keys())), 4, 'fetch tab fetch run list')
    
    # get range
    tab.run.set('40127-40129')
    tab.get_data()
    assert_equal(len(list(tab.data_lines.keys())), 7, 'fetch tab fetch run range')

@with_bfit
def test_fetch_multi_mode(tab=None, b=None):

    # set year
    tab.year.set(2021)
    
    # get runs
    tab.run.set('40124 40131')
    tab.get_data()
    assert_equal(len(list(tab.data_lines.keys())), 2, 'fetch tab fetch 1f and 1x')

@with_bfit    
def test_remove(tab=None, b=None):
    
    # get some data
    tab.year.set(2020)
    tab.run.set('40123-40130')
    tab.get_data()
    
    # remove single
    tab.data_lines['2020.40123'].degrid()
    assert_equal(len(list(tab.data_lines.keys())), 7, 'fetch tab remove single')
    
    # remove all
    tab.remove_all()
    assert_equal(len(list(tab.data_lines.keys())), 0, 'fetch tab remove all')
    
@with_bfit    
def test_draw(tab=None, b=None):
    
    # get some data - bad passing of ax.draw_objs on first obj otherwise??
    tab.year.set(2020)
    tab.run.set('40123')
    tab.get_data()
    tab.draw_all('data', ignore_check=True)
    
    # get some data
    tab.year.set(2020)
    tab.run.set('40123-40126')
    tab.get_data()
    
    # draw stack
    b.draw_style.set('stack')
    tab.draw_all('data', ignore_check=True)
    ax = b.plt.gcf('data').axes[0]

    assert_equal(len(ax.draw_objs), 4, 'fetch tab draw all stack')
    
    tab.run.set('40127-40128')
    tab.get_data()
    tab.draw_all('data', ignore_check=True)
    assert_equal(len(ax.draw_objs), 6, 'fetch tab draw all stack with more data')
    
    # draw new
    b.draw_style.set('new')
    tab.draw_all('data', ignore_check=True)
    assert_equal(len(b.plt.plots['data']), 2, 'fetch tab draw all new')
    
    # draw redraw
    b.draw_style.set('redraw')
    tab.remove_all()
    tab.run.set('40127-40128')
    tab.get_data()
    tab.draw_all('data', ignore_check=True)
    assert_equal(len(plt.gca().draw_objs), 2, 'fetch tab draw all redraw')

@with_bfit    
def test_checkbox(tab=None, b=None):
    
    # get some data
    tab.year.set(2020)
    tab.run.set('40123-40126')
    tab.get_data()
    
    # force check
    tab.check_state.set(False)
    tab.check_all()
    
    assert_equal(all([d.check_state.get() is False for d in tab.data_lines.values()]), True, 'fetch tab force check')
    
    tab.check_state.set(True)
    tab.check_all()
    
    # uncheck one then uncheck data
    tab.data_lines['2020.40123'].check_state.set(False)
    
    tab.check_state_data.set(False)
    tab.check_all_data()
    
    assert_equal(tab.data_lines['2020.40123'].check_data.get(), True, 'fetch tab check data on unchecked item')
    assert_equal(tab.data_lines['2020.40124'].check_data.get(), False, 'fetch tab check data on checked item')
    
    # test check toggle
    tab.toggle_all()
    assert_equal(tab.data_lines['2020.40123'].check_state.get(), True, 'fetch tab toggle check False -> True')
    assert_equal(tab.data_lines['2020.40124'].check_state.get(), False, 'fetch tab toggle check True -> False')

@with_bfit
def test_filter(tab=None, b=None):
    
    # get data
    tab.year.set(2020)
    tab.run.set('40123-40128')
    tab.get_data()
    
    # ensure all runs are activated
    for k in '345678':
        b.data['2020.4012'+k].check_state.set(True)
    
    # deactivate a run which will be later activated
    b.data['2020.40124'].check_state.set(False)
    
    # filter by activation
    tab.text_filter.insert('1.0', '13 < BIAS < 20\nTEMP>281')
    tab.filter_opt.set('activate')
    tab.filter_runs()
    
    # check activation state
    states = [b.data['2020.4012'+k].check_state.get() for k in '45']
    assert all(states), 'Incorrect runs activated on filter'
    
    states = [b.data['2020.4012'+k].check_state.get() for k in '3678']
    assert not any(states), 'Incorrect runs de-activated on filter'
    
    # filter by removal
    tab.filter_opt.set('remove')
    tab.filter_runs()
    
    remaining = [('2020.4012'+k) in b.data.keys() for k in '45']
    assert all(remaining), "Filter remove didn't keep correct runs"
    
    removed = [('2020.4012'+k) in b.data.keys() for k in '3678']
    assert not any(removed), "Filter remove didn't remove correct runs"
    
@with_bfit
def test_flip(tab=None, b=None):
    
    # get data
    tab.year.set(2020)
    tab.run.set('40123')
    tab.get_data()
    data = b.data['2020.40123']
    
    # check asym c
    asym = data.asym('c')
    assert np.mean(asym[1][:10]) > 0, 'Run incorrect orientation no flip (tuple)'
    
    # flip 
    data.flip_asym.set(True)
    asym = data.asym('c')
    assert np.mean(asym[1][:10]) < 0, 'Run incorrect orientation with flip (tuple)'
    
    # check asym dict
    asym = data.asym()
    assert np.mean(asym['c'][:10]) < 0, 'Run incorrect orientation with flip (dict)'
    
    # flip back
    data.flip_asym.set(False)
    asym = data.asym()
    assert np.mean(asym['c'][:10]) > 0, 'Run incorrect orientation no flip (dict)'
    
    
